import json
import os
import unittest

import openmdao.api as om
import pandas as pd
from openmdao.utils.assert_utils import assert_near_equal

from aviary.subsystems.aerodynamics.gasp_based.gaspaero import CruiseAero, LowSpeedAero
from aviary.variable_info.variables import Aircraft, Dynamic, Mission

here = os.path.abspath(os.path.dirname(__file__))
cruise_data = pd.read_csv(os.path.join(here, "data", "aero_data_cruise.csv"))
ground_data = pd.read_csv(os.path.join(here, "data", "aero_data_ground.csv"))
setup_data = json.load(open(os.path.join(here, "data", "aero_data_setup.json")))


class GASPAeroTest(unittest.TestCase):
    """Test overall static and dynamic aero systems in cruise and near-ground flight"""

    cruise_tol = 1.5e-3
    ground_tol = 0.5e-3

    def test_cruise(self):
        prob = om.Problem()
        prob.model.add_subsystem(
            "aero", CruiseAero(num_nodes=2, input_atmos=True), promotes=["*"]
        )
        prob.setup(check=False, force_alloc_complex=True)

        _init_geom(prob)

        # extra params needed for cruise aero
        prob.set_val(Mission.Design.LIFT_COEFFICIENT_MAX_FLAPS_UP, setup_data["clmwfu"])
        prob.set_val(
            Aircraft.Design.SUPERCRITICAL_DIVERGENCE_SHIFT, setup_data["scfac"]
        )

        for i, row in cruise_data.iterrows():
            alt = row["alt"]
            mach = row["mach"]
            alpha = row["alpha"]

            with self.subTest(alt=alt, mach=mach, alpha=alpha):
                # prob.set_val(Dynamic.Mission.ALTITUDE, alt)
                prob.set_val(Dynamic.Mission.MACH, mach)
                prob.set_val("alpha", alpha)
                prob.set_val(Dynamic.Mission.SPEED_OF_SOUND, row["sos"])
                prob.set_val("nu", row["nu"])

                prob.run_model()

                assert_near_equal(prob["CL"][0], row["cl"], tolerance=self.cruise_tol)
                assert_near_equal(prob["CD"][0], row["cd"], tolerance=self.cruise_tol)

    def test_ground(self):
        prob = om.Problem()
        prob.model.add_subsystem(
            "aero", LowSpeedAero(num_nodes=2, input_atmos=True), promotes=["*"]
        )
        prob.setup(check=False, force_alloc_complex=True)

        _init_geom(prob)

        # extra params needed for ground aero
        # leave time ramp inputs as defaults, gear/flaps extended and stay
        prob.set_val(Aircraft.Wing.HEIGHT, 8.0)  # not defined in standalone aero
        prob.set_val("airport_alt", 0.0)  # not defined in standalone aero
        prob.set_val(Aircraft.Wing.FLAP_CHORD_RATIO, setup_data["cfoc"])
        prob.set_val(Mission.Design.GROSS_MASS, setup_data["wgto"])

        for i, row in ground_data.iterrows():
            ilift = row["ilift"]  # 2: takeoff, 3: landing
            alt = row["alt"]
            mach = row["mach"]
            alpha = row["alpha"]

            with self.subTest(ilift=ilift, alt=alt, mach=mach, alpha=alpha):
                prob.set_val(Dynamic.Mission.MACH, mach)
                prob.set_val(Dynamic.Mission.ALTITUDE, alt)
                prob.set_val("alpha", alpha)
                prob.set_val(Dynamic.Mission.SPEED_OF_SOUND, row["sos"])
                prob.set_val("nu", row["nu"])

                # note we're just letting the time ramps for flaps/gear default to the
                # takeoff config such that the default times correspond to full flap and
                # gear increments
                if row["ilift"] == 2:
                    # takeoff flaps config
                    prob.set_val("flap_defl", setup_data["delfto"])
                    prob.set_val("CL_max_flaps", setup_data["clmwto"])
                    prob.set_val("dCL_flaps_model", setup_data["dclto"])
                    prob.set_val("dCD_flaps_model", setup_data["dcdto"])
                    prob.set_val("aero_ramps.flap_factor:final_val", 1.)
                    prob.set_val("aero_ramps.gear_factor:final_val", 1.)
                else:
                    # landing flaps config
                    prob.set_val("flap_defl", setup_data["delfld"])
                    prob.set_val("CL_max_flaps", setup_data["clmwld"])
                    prob.set_val("dCL_flaps_model", setup_data["dclld"])
                    prob.set_val("dCD_flaps_model", setup_data["dcdld"])

                prob.run_model()

                assert_near_equal(prob["CL"][0], row["cl"], tolerance=self.ground_tol)
                assert_near_equal(prob["CD"][0], row["cd"], tolerance=self.ground_tol)

    def test_ground_alpha_out(self):
        """Test that drag output matches between alpha in/out cases"""
        prob = om.Problem()
        prob.model.add_subsystem(
            "alpha_in",
            LowSpeedAero(),
            promotes_inputs=["*", ("alpha", "alpha_in")],
            promotes_outputs=[(Dynamic.Mission.LIFT, "lift_req")],
        )

        prob.model.add_subsystem(
            "alpha_out",
            LowSpeedAero(output_alpha=True),
            promotes_inputs=["*", "lift_req"],
        )

        prob.setup(check=False, force_alloc_complex=True)

        _init_geom(prob)

        # extra params needed for ground aero
        # leave time ramp inputs as defaults, gear/flaps extended and stay
        prob.set_val(Aircraft.Wing.HEIGHT, 8.0)  # not defined in standalone aero
        prob.set_val("airport_alt", 0.0)  # not defined in standalone aero
        prob.set_val(Aircraft.Wing.FLAP_CHORD_RATIO, setup_data["cfoc"])
        prob.set_val(Mission.Design.GROSS_MASS, setup_data["wgto"])

        prob.set_val(Dynamic.Mission.MACH, 0.1)
        prob.set_val(Dynamic.Mission.ALTITUDE, 10)
        prob.set_val("alpha_in", 5)
        prob.run_model()

        assert_near_equal(prob["alpha_in.drag"], prob["alpha_out.drag"], tolerance=1e-6)


def _init_geom(prob):
    """Initialize user inputs and geometry/sizing data"""
    # i.e. common auto IVC vars for the setup + cruise and ground aero models
    prob.set_val(Aircraft.Wing.AREA, setup_data["sw"])
    prob.set_val(Aircraft.Wing.SPAN, setup_data["b"])
    prob.set_val(Aircraft.Wing.AVERAGE_CHORD, setup_data["cbarw"])
    prob.set_val(Aircraft.Wing.TAPER_RATIO, setup_data["slm"])
    prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, setup_data["tcr"])
    prob.set_val(Aircraft.Wing.MOUNTING_TYPE, setup_data["hwing"])
    prob.set_val(Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, setup_data["sah"])
    prob.set_val(Aircraft.HorizontalTail.SPAN, setup_data["bht"])
    prob.set_val(Aircraft.VerticalTail.SPAN, setup_data["bvt"])
    prob.set_val(Aircraft.HorizontalTail.AREA, setup_data["sht"])
    prob.set_val(Aircraft.HorizontalTail.AVERAGE_CHORD, setup_data["cbarht"])
    prob.set_val(Aircraft.Fuselage.AVG_DIAMETER, setup_data["swf"])
    # ground & cruise, dynamic: mach
    prob.set_val(Aircraft.Design.STATIC_MARGIN, setup_data["stmarg"])
    prob.set_val(Aircraft.Design.CG_DELTA, setup_data["delcg"])
    prob.set_val(Aircraft.Wing.ASPECT_RATIO, setup_data["ar"])
    prob.set_val(Aircraft.Wing.SWEEP, setup_data["dlmc4"])
    prob.set_val(Aircraft.HorizontalTail.SWEEP, setup_data["dwpqch"])
    prob.set_val(Aircraft.HorizontalTail.MOMENT_RATIO, setup_data["coelth"])
    # ground & cruise, dynamic: alt
    prob.set_val(Aircraft.Wing.FORM_FACTOR, setup_data["ckw"])
    prob.set_val(Aircraft.Fuselage.FORM_FACTOR, setup_data["ckf"])
    prob.set_val(Aircraft.Nacelle.FORM_FACTOR, setup_data["ckn"])
    prob.set_val(Aircraft.VerticalTail.FORM_FACTOR, setup_data["ckvt"])
    prob.set_val(Aircraft.HorizontalTail.FORM_FACTOR, setup_data["ckht"])
    prob.set_val(Aircraft.Wing.FUSELAGE_INTERFERENCE_FACTOR, setup_data["cki"])
    prob.set_val(Aircraft.Strut.FUSELAGE_INTERFERENCE_FACTOR, setup_data["ckstrt"])
    prob.set_val(Aircraft.Design.DRAG_COEFFICIENT_INCREMENT, setup_data["delcd"])
    prob.set_val(Aircraft.Fuselage.FLAT_PLATE_AREA_INCREMENT, setup_data["delfe"])
    prob.set_val(Aircraft.Wing.CENTER_DISTANCE, setup_data["xwqlf"])
    prob.set_val(Aircraft.Wing.MIN_PRESSURE_LOCATION, setup_data["xcps"])
    prob.set_val(
        Aircraft.Wing.MAX_THICKNESS_LOCATION, 0.4
    )  # overriden in standalone code
    prob.set_val(Aircraft.Strut.AREA_RATIO, setup_data["sstqsw"])
    prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD_TIP, setup_data["tct"])
    prob.set_val(Aircraft.VerticalTail.AVERAGE_CHORD, setup_data["cbarvt"])
    prob.set_val(Aircraft.Fuselage.LENGTH, setup_data["elf"])
    prob.set_val(Aircraft.Nacelle.AVG_LENGTH, setup_data["eln"])
    prob.set_val(Aircraft.Fuselage.WETTED_AREA, setup_data["sf"])
    prob.set_val(Aircraft.Nacelle.SURFACE_AREA, setup_data["sn"] / setup_data["enp"])
    prob.set_val(Aircraft.VerticalTail.AREA, setup_data["svt"])
    prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED, setup_data["tc"])
    prob.set_val(Aircraft.Strut.CHORD, 0)  # not defined in standalone aero
    # ground & cruise, dynamic: alpha
    prob.set_val(Aircraft.Wing.ZERO_LIFT_ANGLE, setup_data["alphl0"])
    # ground & cruise, config-specific: CL_max_flaps
    # ground: wing_height
    # ground: airport_alt
    # ground: flap_defl
    # ground: flap_chord_ratio
    # ground: CL_max_flaps
    # ground: dCL_flaps_model
    # ground: t_init_flaps
    # ground: t_curr
    # ground: dt_flaps
    # ground: gross_mass_initial
    # ground: dCD_flaps_model
    # ground: t_init_gear
    # ground: dt_gear
    # ground & cruise, dynamic: q


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variable_meta_data import _MetaData as _meta_data
from aviary.variable_info.variables import Aircraft, Dynamic


class SimpleDrag(om.ExplicitComponent):
    '''
    Calculate drag as a function of wing area, dynamic pressure, and drag coefficient.

    Apply optional factors (default: 1.0) for increasing or decreasing the drag
    coefficient before calculating drag.
    '''

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        add_aviary_input(self, Aircraft.Wing.AREA, val=1., units='m**2')
        add_aviary_input(self, Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR, val=1.)
        add_aviary_input(self, Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR, val=1.)

        self.add_input(
            Dynamic.Mission.MACH, val=np.ones(nn), units='unitless',
            desc='ratio of local fluid speed to local speed of sound')

        self.add_input(
            Dynamic.Mission.DYNAMIC_PRESSURE, val=np.ones(nn), units='N/m**2',
            desc='pressure caused by fluid motion')

        self.add_input(
            'drag_coefficient', val=np.ones(nn), units='unitless',
            desc='total drag coefficient')

        self.add_output(Dynamic.Mission.DRAG, val=np.ones(nn),
                        units='N', desc='total drag')

    def setup_partials(self):
        nn = self.options['num_nodes']
        rows_cols = np.arange(nn)

        self.declare_partials(
            Dynamic.Mission.DRAG,
            [
                Aircraft.Wing.AREA,
                Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR,
                Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR])

        self.declare_partials(Dynamic.Mission.DRAG,
                              Dynamic.Mission.MACH, dependent=False)

        self.declare_partials(
            Dynamic.Mission.DRAG,
            [Dynamic.Mission.DYNAMIC_PRESSURE, 'drag_coefficient'],
            rows=rows_cols, cols=rows_cols)

    def compute(self, inputs, outputs):
        S = inputs[Aircraft.Wing.AREA]
        FCDSUB = inputs[Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR]
        FCDSUP = inputs[Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR]
        M = inputs[Dynamic.Mission.MACH]
        q = inputs[Dynamic.Mission.DYNAMIC_PRESSURE]
        CD = inputs['drag_coefficient']

        idx_sup = np.where(M >= 1.0)
        CD_scaled = CD * FCDSUB
        CD_scaled[idx_sup] = CD[idx_sup] * FCDSUP

        outputs[Dynamic.Mission.DRAG] = q * S * CD_scaled

    def compute_partials(self, inputs, partials):
        S = inputs[Aircraft.Wing.AREA]
        FCDSUB = inputs[Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR]
        FCDSUP = inputs[Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR]
        M = inputs[Dynamic.Mission.MACH]
        q = inputs[Dynamic.Mission.DYNAMIC_PRESSURE]
        CD = inputs['drag_coefficient']

        idx_sup = np.where(M >= 1.0)
        CD_scaled = CD * FCDSUB
        CD_scaled[idx_sup] = CD[idx_sup] * FCDSUP

        partials[Dynamic.Mission.DRAG, Aircraft.Wing.AREA] = q * CD_scaled
        partials[Dynamic.Mission.DRAG, Dynamic.Mission.DYNAMIC_PRESSURE] = S * CD_scaled

        idx_sub = np.where(M < 1.0)
        dCD = q * S
        dCD[idx_sub] *= FCDSUB
        dCD[idx_sup] *= FCDSUP
        partials[Dynamic.Mission.DRAG, 'drag_coefficient'] = dCD

        drag_unscaled = q * S * CD
        dF = np.zeros(CD.shape)
        dF[idx_sub] = drag_unscaled[idx_sub]
        partials[Dynamic.Mission.DRAG, Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR] = dF

        dF = np.zeros(CD.shape)
        dF[idx_sup] = drag_unscaled[idx_sup]
        partials[Dynamic.Mission.DRAG, Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR] = dF


class TotalDragCoeff(om.ExplicitComponent):
    '''
    Calculate the total drag coefficient from lift-dependent and lift-independent drag
    coefficients.

    Apply an optional factor (default: 1.0) for increasing or decreasing the lift-
    dependent drag coefficient before calculating the total drag coefficient.

    Apply an optional factor (default: 1.0) for increasing or decreasing the lift-
    independent drag coefficient before calculating the total drag coefficient.

    Note, the lift-dependent drag coefficient includes contirbutions from the pressure
    drag coefficient.
    '''

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        add_aviary_input(self, Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR, 1.)
        add_aviary_input(self, Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR, 1.)

        self.add_input(
            'CDI', val=np.ones(nn), units='unitless',
            desc='lift-dependent drag coefficient,'
            ' including contributions from pressure drag coefficient')

        self.add_input(
            'CD0', val=np.ones(nn), units='unitless',
            desc='lift-independent drag coefficient')

        self.add_output(
            'drag_coefficient', val=np.ones(nn), units='unitless',
            desc='total drag coefficient')

    def setup_partials(self):
        nn = self.options['num_nodes']
        rows_cols = np.arange(nn)

        self.declare_partials(
            'drag_coefficient',
            [
                Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR,
                Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR])

        self.declare_partials(
            'drag_coefficient', ['CDI', 'CD0'], rows=rows_cols, cols=rows_cols)

    def compute(self, inputs, outputs):
        FCD0 = inputs[Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR]
        FCDI = inputs[Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR]
        CDI = inputs['CDI']
        CD0 = inputs['CD0']

        outputs['drag_coefficient'] = CDI * FCDI + CD0 * FCD0

    def compute_partials(self, inputs, partials):
        FCD0 = inputs[Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR]
        FCDI = inputs[Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR]
        CDI = inputs['CDI']
        CD0 = inputs['CD0']

        partials['drag_coefficient', Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR] = CD0

        partials[
            'drag_coefficient', Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR] = CDI

        partials['drag_coefficient', 'CDI'] = FCDI
        partials['drag_coefficient', 'CD0'] = FCD0


class TotalDrag(om.Group):
    '''
    Calculate drag as a function of wing area, dynamic pressure, and lift-dependent and
    lift-independent drag coefficients.

    Apply an optional factor (default: 1.0) for increasing or decreasing the lift-
    dependent drag coefficient before calculating the total drag coefficient.

    Apply an optional factor (default: 1.0) for increasing or decreasing the lift-
    independent drag coefficient before calculating the total drag coefficient.

    Note, the lift-dependent drag coefficient includes contirbutions from the pressure
    drag coefficient.

    Apply optional factors (default: 1.0) for increasing or decreasing the total drag
    coefficient before calculating drag. The effect is cummulative with the above
    optional factors.
    '''

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        FCDI_desc = _meta_data[Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR]['desc']
        FCD0_desc = _meta_data[Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR]['desc']

        kwargs = {
            'CDI': dict(
                val=np.ones(nn), units='unitless',
                desc='lift-dependent drag coefficient,'
                ' including contributions from pressure drag coefficient'),
            'CD0': dict(
                val=np.ones(nn), units='unitless',
                desc='lift-independent drag coefficient'),
            'FCDI': dict(val=1., units='unitless', desc=FCDI_desc),
            'FCD0': dict(val=1., units='unitless', desc=FCD0_desc),
            'CD': dict(val=np.ones(nn), units='unitless', desc='total drag coefficient')
        }

        total_drag_comp = self.add_subsystem(
            'total_drag_coeff',
            om.ExecComp('CD = CDI * FCDI + CD0 * FCD0', **kwargs),
            promotes_inputs=[
                'CDI', 'CD0',
                ('FCDI', Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR),
                ('FCD0', Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR)],
            promotes_outputs=[('CD', 'drag_coefficient')])
        total_drag_comp.declare_coloring(show_summary=False)

        self.add_subsystem('simple_drag', SimpleDrag(num_nodes=nn), promotes=['*'])



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft


class TailVolCoef(om.ExplicitComponent):
    """GASP tail volume coefficient fallback calculation.

    This component can be used to compute a volume coefficient for either a horizontal
    or vertical tail. The volume coefficient is based on an empirical relationship
    using gross aircraft parameters such as fuselage length, wing area, etc. For a
    horizontal tail, the wing chord is input. For a vertical tail, the wing span is
    input.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )
        self.options.declare(
            "vertical",
            default=False,
            types=bool,
            desc=(
                "Use vertical tail volume coefficient relationship, "
                "otherwise assume a horizontal tail"
            ),
        )

    def setup(self):
        # coefficients used in the empirical equation
        if self.options["vertical"]:
            self.k = [0.07, 0.0434, 0.336]
        else:
            self.k = [0.43, 0.38, 0.85]

        add_aviary_input(self, Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0)

        add_aviary_input(self, Aircraft.Fuselage.LENGTH, val=129.4)

        self.add_input("cab_w", 13.1, units="ft", desc="SWF: Cabin width")

        add_aviary_input(self, Aircraft.Wing.AREA, val=1370)

        self.add_input(
            "wing_ref",
            12.615,
            units="ft",
            desc=(
                "CBARW | B: Wing reference parameter. Wing chord for a "
                "horizontal tail. Wing span for a vertical tail."
            ),
        )

        self.add_output(
            "vol_coef", units="unitless", desc="VBARH | VBARV: Tail volume coefficient")

    def setup_partials(self):
        self.declare_partials("vol_coef", "*")

    def compute(self, inputs, outputs):
        htail_loc, fus_len, cab_w, wing_area, wing_ref = inputs.values()
        k1, k2, k3 = self.k
        ch1 = k1 - k2 * htail_loc
        outputs["vol_coef"] = k3 * fus_len * cab_w**2 / (wing_area * wing_ref) + ch1

    def compute_partials(self, inputs, J):
        htail_loc, fus_len, cab_w, wing_area, wing_ref = inputs.values()
        k1, k2, k3 = self.k
        J["vol_coef", Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION] = -k2
        J["vol_coef", Aircraft.Fuselage.LENGTH] = (
            k3 * cab_w**2 / (wing_area * wing_ref)
        )
        J["vol_coef", "cab_w"] = 2 * k3 * fus_len * cab_w / (wing_area * wing_ref)
        J["vol_coef", Aircraft.Wing.AREA] = (
            -k3 * fus_len * cab_w**2 / (wing_area**2 * wing_ref)
        )
        J["vol_coef", "wing_ref"] = (
            -k3 * fus_len * cab_w**2 / (wing_area * wing_ref**2)
        )


class TailSize(om.ExplicitComponent):
    """GASP tail geometry calculations.

    This component can be used for either a horizontal tail or vertical tail. For a
    horizontal tail, the ratio of wing chord to tail moment arm and the wing chord are
    input for tail moment arm calculation. For a vertical tail, the ratio of wing span
    to tail moment arm and the wing span are input.
    """

    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):
        # defaults here for Large Single Aisle 1 horizontal tail
        self.add_input(
            "vol_coef", 1.189, units="unitless",
            desc="VBARH | VBARV: Horizontal tail volume coefficient"
        )

        add_aviary_input(self, Aircraft.Wing.AREA, val=1370)

        self.add_input(
            "r_arm",
            0.2307,
            units="unitless",
            desc=(
                "COELTH | BOELTV: For a horizontal tail, the ratio of "
                "wing chord to tail moment arm. For a vertical tail, the "
                "ratio of wing span to vertical tail moment arm."
            ),
        )
        self.add_input(
            "wing_ref",
            12.615,
            units="ft",
            desc=(
                "CBARW | B: Reference wing parameter for tail moment arm. "
                "For a horizontal tail, the mean wing chord. For a "
                "vertical tail, the wing span."
            ),
        )
        self.add_input(
            "ar", 4.75, units="unitless", desc="ARHT | ARVT: Tail aspect ratio.")
        self.add_input(
            "tr", 0.352, units="unitless", desc="SLMH | SLMV: Tail taper ratio.")

        self.add_output("area", units="ft**2", desc="SHT | SVT: Tail area")
        self.add_output("span", units="ft", desc="BHT | BVT: Tail span")
        self.add_output("rchord", units="ft", desc="CRCLHT | CRCLVT: Tail root chord")
        self.add_output(
            "chord", units="ft", desc="CBARHT | CBARVT: Tail mean aerodynamic chord"
        )
        self.add_output("arm", units="ft", desc="ELTH | ELTV: Tail moment arm")

    def setup_partials(self):
        self.declare_partials("area", ["vol_coef", Aircraft.Wing.AREA, "r_arm"])
        self.declare_partials("span", ["vol_coef", Aircraft.Wing.AREA, "r_arm", "ar"])
        self.declare_partials(
            "rchord", ["vol_coef", Aircraft.Wing.AREA, "r_arm", "ar", "tr"]
        )
        self.declare_partials(
            "chord", ["vol_coef", Aircraft.Wing.AREA, "r_arm", "ar", "tr"]
        )
        self.declare_partials("arm", ["r_arm", "wing_ref"])

    def compute(self, inputs, outputs):
        vol_coef, wing_area, r_arm, wing_ref, ar, tr = inputs.values()

        area = vol_coef * wing_area * r_arm
        span = np.sqrt(area * ar)
        rchord = 2 * area / span / (1 + tr)
        chord = 2 / 3.0 * rchord * ((1 + tr) - (tr / (1 + tr)))
        arm = wing_ref / r_arm

        outputs["area"] = area
        outputs["span"] = span
        outputs["rchord"] = rchord
        outputs["chord"] = chord
        outputs["arm"] = arm

    def compute_partials(self, inputs, J):
        vol_coef, wing_area, r_arm, wing_ref, ar, tr = inputs.values()

        J["area", "vol_coef"] = wing_area * r_arm
        J["area", Aircraft.Wing.AREA] = vol_coef * r_arm
        J["area", "r_arm"] = vol_coef * wing_area

        cse1 = np.sqrt(vol_coef * wing_area * r_arm * ar)
        J["span", "vol_coef"] = cse1 / (2 * vol_coef)
        J["span", Aircraft.Wing.AREA] = cse1 / (2 * wing_area)
        J["span", "r_arm"] = cse1 / (2 * r_arm)
        J["span", "ar"] = cse1 / (2 * ar)

        cse2 = cse1 * (tr + 1)
        J["rchord", "vol_coef"] = wing_area * r_arm / cse2
        J["rchord", Aircraft.Wing.AREA] = vol_coef * r_arm / cse2
        J["rchord", "r_arm"] = wing_area * vol_coef / cse2
        J["rchord", "ar"] = -vol_coef * wing_area * r_arm / (ar * cse2)
        J["rchord", "tr"] = -2 * vol_coef * wing_area * r_arm / (cse2 * (tr + 1))

        cse3 = tr - (tr / (tr + 1)) + 1
        J["chord", "vol_coef"] = 2 / 3.0 * wing_area * r_arm * cse3 / cse2
        J["chord", Aircraft.Wing.AREA] = 2 / 3.0 * vol_coef * r_arm * cse3 / cse2
        J["chord", "r_arm"] = 2 / 3.0 * vol_coef * wing_area * cse3 / cse2
        J["chord", "ar"] = -2 / 3.0 * vol_coef * wing_area * r_arm * cse3 / (ar * cse2)
        J["chord", "tr"] = 4 / 3.0 * cse1 * (tr - 1) / (ar * (tr + 1) ** 3)

        J["arm", "r_arm"] = -wing_ref / r_arm**2
        J["arm", "wing_ref"] = 1.0 / r_arm


class EmpennageSize(om.Group):
    """GASP geometry calculations for both horizontal and vertical tails.

    Volume coefficients for the tails may be either specified directly (default) or
    computed via empirical relationships to general airplane parameters.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):
        # TODO: For cruciform/T-tail configurations, GASP checks to make sure the V tail
        # chord at the H tail location is greater than the H tail root chord. If not, it
        # overrides the H tail taper ratio so they match. If that leads to a H tail root
        # chord greater than the H tail tip chord, it sets the taper ratio to 1 and
        # overrides the H tail aspect ratio. H tail taper ratio is used in landing gear
        # mass calculation.

        aviary_options = self.options['aviary_options']

        # higher inputs that are input to groups other than this one, or calculated in groups other than this one
        higher_level_inputs_htail_vc = [
            ("cab_w", Aircraft.Fuselage.AVG_DIAMETER),
            ("wing_ref", Aircraft.Wing.AVERAGE_CHORD),
        ]
        higher_level_inputs_vtail_vc = [
            ("cab_w", Aircraft.Fuselage.AVG_DIAMETER),
            ("wing_ref", Aircraft.Wing.SPAN),
        ]
        higher_level_inputs_htail = [
            ("wing_ref", Aircraft.Wing.AVERAGE_CHORD),
            ("tr", Aircraft.HorizontalTail.TAPER_RATIO),
        ]
        higher_level_inputs_vtail = [
            ("wing_ref", Aircraft.Wing.SPAN),
            ("ar", Aircraft.VerticalTail.ASPECT_RATIO),
        ]

        # inputs to htail and vtail that are purely promoted to change the name
        rename_inputs_htail = [
            ("r_arm", Aircraft.HorizontalTail.MOMENT_RATIO),
            ("ar", Aircraft.HorizontalTail.ASPECT_RATIO),
        ]
        rename_inputs_vtail = [
            ("r_arm", Aircraft.VerticalTail.MOMENT_RATIO),
            ("tr", Aircraft.VerticalTail.TAPER_RATIO),
        ]

        # outputs that are used in groups other than this one
        higher_level_outputs_htail = [
            ("area", Aircraft.HorizontalTail.AREA),
            ("span", Aircraft.HorizontalTail.SPAN),
            ("rchord", Aircraft.HorizontalTail.ROOT_CHORD),
            ("arm", Aircraft.HorizontalTail.MOMENT_ARM),
        ]
        higher_level_outputs_vtail = [
            ("area", Aircraft.VerticalTail.AREA),
            ("span", Aircraft.VerticalTail.SPAN),
            ("rchord", Aircraft.VerticalTail.ROOT_CHORD),
            ("arm", Aircraft.VerticalTail.MOMENT_ARM),
        ]

        # outputs from htail and vtail that are purely promoted to change the name
        rename_outputs_htail = [
            ("chord", Aircraft.HorizontalTail.AVERAGE_CHORD),
        ]
        rename_outputs_vtail = [
            ("chord", Aircraft.VerticalTail.AVERAGE_CHORD),
        ]

        # outputs from components in this group that are used in this group
        connected_outputs_htail_vc = [
            ("vol_coef", Aircraft.HorizontalTail.VOLUME_COEFFICIENT),
        ]
        connected_outputs_vtail_vc = [
            ("vol_coef", Aircraft.VerticalTail.VOLUME_COEFFICIENT),
        ]

        # inputs to components in this group that are calculated in this group
        connected_inputs_htail = [
            ("vol_coef", Aircraft.HorizontalTail.VOLUME_COEFFICIENT),
        ]
        connected_inputs_vtail = [
            ("vol_coef", Aircraft.VerticalTail.VOLUME_COEFFICIENT),
        ]

        if self.options["aviary_options"].get_val(Aircraft.Design.COMPUTE_HTAIL_VOLUME_COEFF, units='unitless'):
            self.add_subsystem(
                "htail_vc",
                TailVolCoef(aviary_options=aviary_options),
                promotes_inputs=higher_level_inputs_htail_vc + ["aircraft:*"],
                promotes_outputs=connected_outputs_htail_vc,
            )
        if self.options["aviary_options"].get_val(Aircraft.Design.COMPUTE_VTAIL_VOLUME_COEFF, units='unitless'):
            self.add_subsystem(
                "vtail_vc",
                TailVolCoef(aviary_options=aviary_options, vertical=True),
                promotes_inputs=higher_level_inputs_vtail_vc + ["aircraft:*"],
                promotes_outputs=connected_outputs_vtail_vc,
            )

        self.add_subsystem(
            "htail",
            TailSize(aviary_options=aviary_options,),
            promotes_inputs=higher_level_inputs_htail
            + rename_inputs_htail
            + connected_inputs_htail
            + ["aircraft:*"],
            promotes_outputs=higher_level_outputs_htail + rename_outputs_htail,
        )

        self.add_subsystem(
            "vtail",
            TailSize(aviary_options=aviary_options,),
            promotes_inputs=higher_level_inputs_vtail
            + rename_inputs_vtail
            + connected_inputs_vtail
            + ["aircraft:*"],
            promotes_outputs=higher_level_outputs_vtail + rename_outputs_vtail,
        )

        self.set_input_defaults(Aircraft.Wing.AVERAGE_CHORD, 12.615, units="ft")
        self.set_input_defaults(Aircraft.Wing.SPAN, 117.8054, units="ft")

        # override horizontal tail defaults
        self.set_input_defaults(Aircraft.VerticalTail.VOLUME_COEFFICIENT, 0.145)
        self.set_input_defaults(Aircraft.VerticalTail.MOMENT_RATIO, 2.362)
        self.set_input_defaults(Aircraft.VerticalTail.ASPECT_RATIO, 1.67)
        self.set_input_defaults(Aircraft.VerticalTail.TAPER_RATIO, 0.801)



"""
This module, process_input_decks.py, is responsible for reading vehicle input decks, initializing options, 
and setting initial guesses for aircraft design parameters. It works primarily with .csv files,
allowing for the specification of units, comments, and lists within these files. 

The module supports various functions like creating a vehicle, parsing input files, updating options based
on inputs, and handling initial guesses for different aircraft design aspects. It heavily relies on the 
aviary and openMDAO libraries for processing and interpreting the aircraft design parameters.

Functions:
    create_vehicle(vehicle_deck=''): Create and initialize a vehicle with default or specified parameters.
    parse_inputs(vehicle_deck, aircraft_values): Parse input files and update aircraft values and initial guesses.
    update_options(aircraft_values, initial_guesses): Update dependent options based on current aircraft values.
    update_dependent_options(aircraft_values, dependent_options): Update options that depend on the value of an input variable.
    initial_guessing(aircraft_values): Set initial guesses for aircraft parameters based on problem type and other factors.
"""

import warnings
from operator import eq, ge, gt, le, lt, ne

import numpy as np
from openmdao.utils.units import valid_units

from aviary.utils.aviary_values import AviaryValues, get_keys
from aviary.utils.functions import convert_strings_to_data, set_value
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.enums import ProblemType
from aviary.variable_info.variable_meta_data import _MetaData
from aviary.variable_info.variables import Aircraft, Mission
from aviary.utils.functions import get_path


operation_dict = {"<": lt, "<=": le, "==": eq, "!=": ne,
                  ">=": ge, ">": gt, "isinstance": isinstance}
problem_types = {'sizing': ProblemType.SIZING,
                 'alternate': ProblemType.ALTERNATE, 'fallout': ProblemType.FALLOUT}


def create_vehicle(vehicle_deck=''):
    """
    Creates and initializes a vehicle with default or specified parameters. It sets up the aircraft values
    and initial guesses based on the input from the vehicle deck.

    Parameters
    ----------
    vehicle_deck (str): Path to the vehicle deck file. Default is an empty string.

    Returns
    -------
    tuple: Returns a tuple containing aircraft values and initial guesses.
    """
    aircraft_values = get_option_defaults(engine=False)

    # TODO temporary, needed until debug_mode retired in favor of new verbosity flag
    # TODO remove all hardcoded GASP values here, find appropriate place for them
    aircraft_values.set_val('debug_mode', val=False)
    aircraft_values.set_val('INGASP.JENGSZ', val=4)
    aircraft_values.set_val('test_mode', val=False)
    aircraft_values.set_val('use_surrogates', val=True)
    aircraft_values.set_val('mass_defect', val=10000, units='lbm')
    aircraft_values.set_val('problem_type', val=ProblemType.SIZING)
    aircraft_values.set_val(Aircraft.Electrical.HAS_HYBRID_SYSTEM, val=False)

    if isinstance(vehicle_deck, AviaryValues):
        aircraft_values.update(vehicle_deck)
        initial_guesses = {}
    else:
        vehicle_deck = get_path(vehicle_deck)
        aircraft_values, initial_guesses = parse_inputs(vehicle_deck, aircraft_values)

    return aircraft_values, initial_guesses


def parse_inputs(vehicle_deck, aircraft_values: AviaryValues(), meta_data=_MetaData):
    """
    Parses the input files and updates the aircraft values and initial guesses. The function reads the
    vehicle deck file, processes each line, and updates the aircraft_values object based on the data found.

    Parameters
    ----------
    vehicle_deck (str): The vehicle deck file path.
    aircraft_values (AviaryValues): An instance of AviaryValues to be updated.

    Returns
    -------
    tuple: Updated aircraft values and initial guesses.
    """
    initial_guesses = {
        # initial_guesses is a dictionary that contains values used to initialize the trajectory
        'actual_takeoff_mass': 0,
        'rotation_mass': .99,
        'operating_empty_mass': 0,
        'fuel_burn_per_passenger_mile': 0.1,
        'cruise_mass_final': 0,
        'flight_duration': 0,
        'time_to_climb': 0,
        'climb_range': 0,
        'reserves': 0
    }
    guess_names = list(initial_guesses.keys())

    with open(vehicle_deck, newline='') as f_in:
        for line in f_in:
            used, data_units = False, None

            tmp = [*line.split('#', 1), '']
            line, comment = tmp[0], tmp[1]  # anything after the first # is a comment

            data = ''.join(line.rstrip(',').split())  # remove all white space

            if len(data) == 0:
                continue  # skip line it contained only commas

            # remove any elements that are empty (caused by trailing commas or extra commas)
            data_list = [dat for dat in data.split(',') if dat != '']

            # continue if there's no data in the line but there are commas
            # this might occur if someone edits a .csv file in Excel
            if len(data_list) == 0:
                continue
            var_name = data_list.pop(0)
            if valid_units(data_list[-1]):
                # if the last element is a unit, remove it from the list and update the variable's units
                data_units = data_list.pop()

            is_array = False
            if '[' in data_list[0]:
                is_array = True

            var_values = convert_strings_to_data(data_list)

            if var_name == 'debug_mode':
                aircraft_values = set_value(var_name, var_values, aircraft_values)
                continue

            if var_name in meta_data.keys():
                aircraft_values = set_value(
                    var_name, var_values, aircraft_values, units=data_units, is_array=is_array, meta_data=meta_data)
                continue

            elif var_name in guess_names:
                # all initial guesses take only a single value
                initial_guesses[var_name] = float(var_values[0])
                continue

            if 'debug_mode' in aircraft_values and aircraft_values.get_val('debug_mode'):
                print('Unused:', var_name, var_values, comment)

    return aircraft_values, initial_guesses

# TODO this should be a preprocessor, and tasks split to be specific to subsystem
#      e.g. aero preprocessor, mass preprocessor, 2DOF preprocessor, etc.


def update_GASP_options(aircraft_values: AviaryValues()):
    """
    Updates options based on the current values in aircraft_values. This function also handles special cases 
    and prints debug information if the debug mode is active.

    Parameters
    ----------
    aircraft_values (AviaryValues): An instance of AviaryValues containing current aircraft values.

    Returns
    -------
    tuple: Updated aircraft values and initial guesses.
    """
    # update the options that depend on variables
    update_dependent_options(aircraft_values, dependent_options)

    ## STRUT AND FOLD ##
    if not aircraft_values.get_val(Aircraft.Wing.HAS_STRUT):
        aircraft_values.set_val(
            Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED, val=False)

    if aircraft_values.get_val(Aircraft.Wing.HAS_FOLD):
        if not aircraft_values.get_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION):
            aircraft_values.set_val(
                Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED, val=True)
        else:
            dim_loc_spec = aircraft_values.get_val(
                Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED)
            aircraft_values.set_val(
                Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED, val=dim_loc_spec)
    else:
        aircraft_values.set_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, val=True)
        aircraft_values.set_val(
            Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED, val=False)

    if aircraft_values.get_val('debug_mode'):
        print('\nOptions')
        for key in get_keys(aircraft_values):
            val, units = aircraft_values.get_item(key)
            print(key, val, units)

    return aircraft_values


def update_dependent_options(aircraft_values: AviaryValues(), dependent_options):
    """
    Updates options that are dependent on the value of an input variable or option. The function iterates 
    through each dependent option and sets its value based on the current aircraft values.

    Parameters
    ----------
    aircraft_values (AviaryValues): An instance of AviaryValues containing current aircraft values.
    dependent_options (list): A list of dependent options and their dependencies.

    Returns
    -------
    AviaryValues: Updated aircraft values with modified dependent options.
    """
    # gets the names of all the variables that affect dependent options
    for var_name, dependency in dependent_options:
        if var_name in get_keys(aircraft_values):
            var_value, var_units = aircraft_values.get_item(var_name)
            # dependency is a dictionary that contains the target option, the relationship to the variable and the output values
            if dependency['relation'] in operation_dict:
                comp = operation_dict[dependency['relation']]
                outcome = dependency['result'] if comp(
                    var_value, dependency['val']) else dependency['alternate']
            elif dependency['relation'] == "in":
                outcome = dependency['result'] if var_value in dependency['val'] else dependency['alternate']
            else:
                warnings.warn(dependency['relation'] +
                              ' is not a valid selection')
            aircraft_values.set_val(dependency['target'], val=outcome)
    return aircraft_values


def initial_guessing(aircraft_values: AviaryValues(), initial_guesses):
    """
    Sets initial guesses for various aircraft parameters based on the current problem type, aircraft values,
    and other factors. It calculates and sets values like takeoff mass, cruise mass, flight duration, etc.

    Parameters
    ----------
    aircraft_values (AviaryValues): An instance of AviaryValues containing current aircraft values.

    Returns
    -------
    tuple: Updated aircraft values and initial guesses.
    """
    problem_type = aircraft_values.get_val('problem_type')
    num_pax = aircraft_values.get_val(Aircraft.CrewPayload.NUM_PASSENGERS)
    reserve_val = aircraft_values.get_val(
        Aircraft.Design.RESERVE_FUEL_ADDITIONAL, units='lbm')
    reserve_frac = aircraft_values.get_val(
        Aircraft.Design.RESERVE_FUEL_FRACTION, units='unitless')

    reserves = initial_guesses['reserves']
    if reserves < 0.0:
        raise ValueError(
            'initial_guesses["reserves"] must be greater than or equal to 0.')
    elif reserves == 0:
        reserves += reserve_val
        reserves += (reserve_frac * (num_pax * initial_guesses['fuel_burn_per_passenger_mile'] *
                                     aircraft_values.get_val(Mission.Design.RANGE, units='NM')))
    elif reserves < 10:
        reserves *= (num_pax * initial_guesses['fuel_burn_per_passenger_mile'] *
                     aircraft_values.get_val(Mission.Design.RANGE, units='NM'))

    initial_guesses['reserves'] = reserves

    if Mission.Summary.GROSS_MASS in aircraft_values:
        mission_mass = aircraft_values.get_val(Mission.Summary.GROSS_MASS, units='lbm')
    else:
        mission_mass = aircraft_values.get_val(Mission.Design.GROSS_MASS, units='lbm')

    if Mission.Summary.CRUISE_MASS_FINAL in aircraft_values:
        cruise_mass_final = aircraft_values.get_val(
            Mission.Summary.CRUISE_MASS_FINAL, units='lbm')
    else:
        cruise_mass_final = initial_guesses['cruise_mass_final']

    # takeoff mass not given
    if mission_mass <= 0:
        if problem_type == ProblemType.ALTERNATE:
            fuel_mass = num_pax * (
                initial_guesses['fuel_burn_per_passenger_mile'] * aircraft_values.get_val(Mission.Design.RANGE, units='NM')) + reserves
            mission_mass = initial_guesses['operating_empty_mass'] + (
                num_pax * aircraft_values.get_val(Aircraft.CrewPayload.PASSENGER_MASS_WITH_BAGS, units='lbm')) + fuel_mass
        elif problem_type == ProblemType.FALLOUT or problem_type == ProblemType.SIZING:
            mission_mass = aircraft_values.get_val(
                Mission.Design.GROSS_MASS, units='lbm')
    initial_guesses['actual_takeoff_mass'] = mission_mass

    if cruise_mass_final == 0:  # no guess given
        if problem_type == ProblemType.SIZING:
            cruise_mass_final = .8
        elif problem_type == ProblemType.ALTERNATE:
            cruise_mass_final = -1
    # estimation based on payload and fuel
    if cruise_mass_final <= 0:
        cruise_mass_final = initial_guesses['operating_empty_mass'] + \
            num_pax * \
            aircraft_values.get_val(
                Aircraft.CrewPayload.PASSENGER_MASS_WITH_BAGS, units='lbm') + reserves
    # fraction of takeoff mass
    elif cruise_mass_final <= 1:
        cruise_mass_final = mission_mass * \
            cruise_mass_final
    initial_guesses['cruise_mass_final'] = cruise_mass_final

    if initial_guesses['rotation_mass'] <= 1:  # fraction of takeoff mass
        initial_guesses['rotation_mass'] = mission_mass * \
            initial_guesses['rotation_mass']

    if Mission.Design.MACH in aircraft_values:
        cruise_mach = aircraft_values.get_val(Mission.Design.MACH)
    else:
        cruise_mach = aircraft_values.get_val(Mission.Summary.CRUISE_MACH)

    if initial_guesses['flight_duration'] <= 0:  # estimation based on mach
        initial_guesses['flight_duration'] = aircraft_values.get_val(
            Mission.Design.RANGE, units='NM') / (667 * cruise_mach) * (60 * 60)
    elif initial_guesses['flight_duration'] <= 15:  # duration entered in hours
        initial_guesses['flight_duration'] = initial_guesses['flight_duration'] * \
            (60 * 60)

    total_thrust = aircraft_values.get_val(
        Aircraft.Engine.SCALED_SLS_THRUST, 'lbf') * aircraft_values.get_val(Aircraft.Engine.NUM_ENGINES)
    gamma_guess = np.arcsin(.5*total_thrust / mission_mass)
    avg_speed_guess = (.5 * 667 * cruise_mach)  # kts

    if initial_guesses['time_to_climb'] <= 0:  # no guess given
        initial_guesses['time_to_climb'] = aircraft_values.get_val(Mission.Design.CRUISE_ALTITUDE, units='ft') / \
            (avg_speed_guess * np.sin(gamma_guess))
    elif initial_guesses['time_to_climb'] <= 2:  # duration entered in hours
        initial_guesses['time_to_climb'] = initial_guesses['time_to_climb'] * (60 * 60)
    elif initial_guesses['time_to_climb'] <= 200:  # average climb rate in ft/s
        initial_guesses['time_to_climb'] = aircraft_values.get_val(Mission.Design.CRUISE_ALTITUDE, units='ft') / \
            initial_guesses['time_to_climb']

    # range covered using an average speed from 0 to cruise
    if initial_guesses['climb_range'] <= 0:
        initial_guesses['climb_range'] = initial_guesses['time_to_climb'] / \
            (60 * 60) * (avg_speed_guess * np.cos(gamma_guess))

    if aircraft_values.get_val('debug_mode'):
        print('\nInitial Guesses')
        for key, value in initial_guesses.items():
            print(key, value)

    return initial_guesses


dependent_options = [
    # dependent_options is a list that is used to update options that depend on the value of
    # an input variable or option.
    # Each dependency comes in the form of a list with the variable name and a dictionary.
    # The dictionary contains a value that the variable will be compared against (val), the
    # particular mathematical comparison (relation), the option that is effected (target), and
    # the value the option should be set to based on whether the relation is true (result) or
    # false (alternate)
    # For example, an aircraft's engines are on the fuselage if the span fraction on the wing is 0.
    # In this case the value of Aircraft.Engine.WING_LOCATIONS is compared to 0. If the span location
    # is exactly equal to zero, engine_on_fuselage is set to to True, otherwise it is set to False.
    # One variable can be used to set the value of any number of options, but an option can only be
    # set by one variable

    # [Aircraft.Engine.WING_LOCATIONS, {
    #     'val': 0, 'relation': '==', 'target': Aircraft.Engine.FUSELAGE_MOUNTED, 'result': True, 'alternate': False}],
    [Aircraft.Fuselage.WETTED_AREA_FACTOR, {
        'val': 10, 'relation': '>', 'target': Aircraft.Fuselage.PROVIDE_SURFACE_AREA, 'result': True, 'alternate': False}],
    [Aircraft.Wing.LOADING, {'val': 20, 'relation': '>',
                             'target': Aircraft.Wing.LOADING_ABOVE_20, 'result': True, 'alternate': False}],
    [Aircraft.Strut.ATTACHMENT_LOCATION, {
        'val': 0, 'relation': '!=', 'target': Aircraft.Wing.HAS_STRUT, 'result': True, 'alternate': False}],
    [Aircraft.Strut.ATTACHMENT_LOCATION, {
        'val': 1, 'relation': '>', 'target': Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED, 'result': True, 'alternate': False}],
    [Aircraft.Wing.FOLD_MASS_COEFFICIENT, {
        'val': 0, 'relation': '>', 'target': Aircraft.Wing.HAS_FOLD, 'result': True, 'alternate': False}],
    [Aircraft.Wing.FOLDED_SPAN, {'val': 1, 'relation': '>', 'target': Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                                 'result': True, 'alternate': False}],
    [Aircraft.Design.PART25_STRUCTURAL_CATEGORY, {
        'val': 0, 'relation': '<', 'target': Aircraft.Design.ULF_CALCULATED_FROM_MANEUVER, 'result': True, 'alternate': False}],
    [Aircraft.Engine.TYPE, {
        'val': [1, 2, 3, 4, 11, 12, 13, 14], 'relation': 'in', 'target': Aircraft.Engine.HAS_PROPELLERS, 'result': True, 'alternate': False}],
    ['JENGSZ', {
        'val': 4, 'relation': '!=', 'target': Aircraft.Engine.SCALE_PERFORMANCE, 'result': True, 'alternate': False}],
    [Aircraft.HorizontalTail.VOLUME_COEFFICIENT, {
        'val': 0, 'relation': '==', 'target': Aircraft.Design.COMPUTE_HTAIL_VOLUME_COEFF, 'result': True, 'alternate': False}],
    [Aircraft.VerticalTail.VOLUME_COEFFICIENT, {
        'val': 0, 'relation': '==', 'target': Aircraft.Design.COMPUTE_VTAIL_VOLUME_COEFF, 'result': True, 'alternate': False}],
]



import unittest
from parameterized import parameterized

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials

from aviary.subsystems.premission import CorePreMission
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.aviary_values import get_items
from aviary.validation_cases.validation_tests import (
    flops_validation_test, get_flops_inputs, get_flops_outputs, get_flops_case_names, print_case
)
from aviary.variable_info.variables import Aircraft, Mission
from aviary.variable_info.variables_in import VariablesIn
from aviary.utils.functions import set_aviary_initial_values
from aviary.interface.default_phase_info.height_energy import default_premission_subsystems
from aviary.utils.preprocessors import preprocess_crewpayload


class PreMissionGroupTest(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):
        flops_inputs = get_flops_inputs(case_name)
        flops_outputs = get_flops_outputs(case_name)
        flops_inputs.set_val(Aircraft.Propulsion.TOTAL_NUM_WING_ENGINES,
                             flops_outputs.get_val(Aircraft.Propulsion.TOTAL_NUM_WING_ENGINES))

        preprocess_crewpayload(flops_inputs)

        prob = self.prob

        prob.model.add_subsystem(
            "pre_mission",
            CorePreMission(aviary_options=flops_inputs,
                           subsystems=default_premission_subsystems),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.model.add_subsystem(
            'input_sink',
            VariablesIn(aviary_options=flops_inputs),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        set_aviary_initial_values(prob.model, flops_inputs)
        prob.setup(check=False, force_alloc_complex=True)
        prob.set_solver_print(2)

        # Initial guess for gross weight.
        # We set it to an unconverged value to test convergence.
        prob.set_val(Mission.Design.GROSS_MASS, val=1000.0)

        # Set inital values for all variables.
        for (key, (val, units)) in get_items(flops_inputs):
            try:
                prob.set_val(key, val, units)

            except KeyError:
                # This is an option, not a variable
                continue

        if case_name in ['LargeSingleAisle1FLOPS', 'LargeSingleAisle2FLOPSdw']:
            # We set these so that their derivatives are defined.
            # The ref values are not set in our test models.
            prob[Aircraft.Wing.ASPECT_RATIO_REF] = prob[Aircraft.Wing.ASPECT_RATIO]
            prob[Aircraft.Wing.THICKNESS_TO_CHORD_REF] = prob[Aircraft.Wing.THICKNESS_TO_CHORD]

        prob[Aircraft.Propulsion.TOTAL_ENGINE_MASS] = flops_outputs.get_val(
            Aircraft.Propulsion.TOTAL_ENGINE_MASS, units='lbm')
        prob[Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS] = flops_outputs.get_val(
            Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS, units='lbm')

        flops_validation_test(
            prob,
            case_name,
            input_keys=[],
            output_keys=[Aircraft.Design.STRUCTURE_MASS,
                         Aircraft.Propulsion.MASS,
                         Aircraft.Design.SYSTEMS_EQUIP_MASS,
                         Aircraft.Design.EMPTY_MASS,
                         Aircraft.Design.OPERATING_MASS,
                         Aircraft.Design.ZERO_FUEL_MASS,
                         Mission.Design.FUEL_MASS],
            step=1.01e-40,
            atol=1e-8,
            rtol=1e-10)

    def test_diff_configuration_mass(self):
        # This standalone test provides coverage for some features unique to this
        # model.
        from aviary.models.large_single_aisle_2.large_single_aisle_2_FLOPS_data import LargeSingleAisle2FLOPS

        prob = om.Problem()

        flops_inputs: AviaryValues = LargeSingleAisle2FLOPS['inputs']
        flops_outputs: AviaryValues = LargeSingleAisle2FLOPS['outputs']

        prob.model.add_subsystem(
            "pre_mission",
            CorePreMission(aviary_options=flops_inputs,
                           subsystems=default_premission_subsystems),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.model.add_subsystem(
            'input_sink',
            VariablesIn(aviary_options=flops_inputs),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        set_aviary_initial_values(prob.model, flops_inputs)
        prob.setup(check=False)

        # Set inital values for all variables.
        for (key, (val, units)) in get_items(flops_inputs):
            try:
                prob.set_val(key, val, units)

            except KeyError:
                # This is an option, not a variable
                continue

        prob.run_model()

        flops_validation_test(
            prob,
            "LargeSingleAisle2FLOPS",
            input_keys=[],
            output_keys=[Mission.Design.FUEL_MASS],
            atol=1e-4,
            rtol=1e-4,
            check_partials=False,
            flops_inputs=flops_inputs,
            flops_outputs=flops_outputs)


if __name__ == "__main__":
    unittest.main()



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.wing_common import (
    WingBendingMass, WingMiscMass, WingShearControlMass)
from aviary.variable_info.options import get_option_defaults
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (flops_validation_test,
                                                      get_flops_case_names,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Mission


class WingShearControlMassTest(unittest.TestCase):
    def setUp(self):
        prob = self.prob = om.Problem()
        prob.model.add_subsystem(
            "wing",
            WingShearControlMass(),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Wing.COMPOSITE_FRACTION,
                        Aircraft.Wing.CONTROL_SURFACE_AREA,
                        Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER,
                        Mission.Design.GROSS_MASS],
            output_keys=Aircraft.Wing.SHEAR_CONTROL_MASS,
            atol=1e-11,
            rtol=1e-11)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class WingMiscMassTest(unittest.TestCase):
    def setUp(self):
        prob = self.prob = om.Problem()
        prob.model.add_subsystem(
            "wing",
            WingMiscMass(),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Wing.COMPOSITE_FRACTION,
                        Aircraft.Wing.AREA,
                        Aircraft.Wing.MISC_MASS_SCALER],
            output_keys=Aircraft.Wing.MISC_MASS,
            atol=1e-11,
            rtol=1e-11)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class WingBendingMassTest(unittest.TestCase):
    def setUp(self):
        prob = self.prob = om.Problem()
        prob.model.add_subsystem(
            "wing",
            WingBendingMass(aviary_options=get_option_defaults()),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR,
                        Aircraft.Wing.BENDING_FACTOR,
                        Aircraft.Wing.BENDING_MASS_SCALER,
                        Aircraft.Wing.COMPOSITE_FRACTION,
                        Aircraft.Wing.ENG_POD_INERTIA_FACTOR,
                        Mission.Design.GROSS_MASS,
                        Aircraft.Wing.LOAD_FRACTION,
                        Aircraft.Wing.MISC_MASS,
                        Aircraft.Wing.MISC_MASS_SCALER,
                        Aircraft.Wing.SHEAR_CONTROL_MASS,
                        Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER,
                        Aircraft.Wing.SPAN,
                        Aircraft.Wing.SWEEP,
                        Aircraft.Wing.ULTIMATE_LOAD_FACTOR,
                        Aircraft.Wing.VAR_SWEEP_MASS_PENALTY],
            output_keys=Aircraft.Wing.BENDING_MASS,
            atol=1e-11,
            rtol=1e-11)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



import numpy as np

from aviary.mission.phase_builder_base import PhaseBuilderBase
from aviary.mission.initial_guess_builders import InitialGuessState, InitialGuessTime, InitialGuessControl
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Dynamic
from aviary.mission.gasp_based.ode.rotation_ode import RotationODE


class RotationPhase(PhaseBuilderBase):
    default_name = 'rotation_phase'
    default_ode_class = RotationODE

    _meta_data_ = {}
    _initial_guesses_meta_data_ = {}

    def build_phase(self, aviary_options: AviaryValues = None):
        phase = self.phase = super().build_phase(aviary_options)

        # Retrieve user options values
        user_options = self.user_options
        fix_initial = user_options.get_val('fix_initial')
        angle_lower = user_options.get_val('angle_lower', units='rad')
        angle_upper = user_options.get_val('angle_upper', units='rad')
        angle_ref = user_options.get_val('angle_ref', units='rad')
        angle_ref0 = user_options.get_val('angle_ref0', units='rad')
        angle_defect_ref = user_options.get_val('angle_defect_ref', units='rad')
        distance_lower = user_options.get_val('distance_lower', units='ft')
        distance_upper = user_options.get_val('distance_upper', units='ft')
        distance_ref = user_options.get_val('distance_ref', units='ft')
        distance_ref0 = user_options.get_val('distance_ref0', units='ft')
        distance_defect_ref = user_options.get_val('distance_defect_ref', units='ft')
        normal_ref = user_options.get_val('normal_ref', units='lbf')
        normal_ref0 = user_options.get_val('normal_ref0', units='lbf')

        # Add states
        phase.add_state(
            "alpha",
            fix_initial=True,
            fix_final=False,
            lower=angle_lower,
            upper=angle_upper,
            units="rad",
            rate_source="alpha_rate",
            ref=angle_ref,
            ref0=angle_ref0,
            defect_ref=angle_defect_ref,
        )

        self.add_velocity_state(user_options)

        self.add_mass_state(user_options)

        phase.add_state(
            Dynamic.Mission.DISTANCE,
            fix_initial=fix_initial,
            fix_final=False,
            lower=distance_lower,
            upper=distance_upper,
            units="ft",
            rate_source="distance_rate",
            ref=distance_ref,
            ref0=distance_ref0,
            defect_ref=distance_defect_ref,
        )

        # Add parameters
        phase.add_parameter("t_init_gear", units="s",
                            static_target=True, opt=False, val=100)
        phase.add_parameter("t_init_flaps", units="s",
                            static_target=True, opt=False, val=100)

        # Add boundary constraints
        phase.add_boundary_constraint(
            "normal_force",
            loc="final",
            equals=0,
            units="lbf",
            ref=normal_ref,
            ref0=normal_ref0,
        )

        # Add timeseries outputs
        phase.add_timeseries_output(Dynamic.Mission.THRUST_TOTAL, units="lbf")
        phase.add_timeseries_output("normal_force")
        phase.add_timeseries_output(Dynamic.Mission.MACH)
        phase.add_timeseries_output("EAS", units="kn")
        phase.add_timeseries_output(Dynamic.Mission.LIFT)
        phase.add_timeseries_output("CL")
        phase.add_timeseries_output("CD")
        phase.add_timeseries_output("fuselage_pitch", output_name="theta", units="deg")

        return phase


# Adding metadata for the RotationPhase
RotationPhase._add_meta_data('fix_initial', val=False)
RotationPhase._add_meta_data('duration_bounds', val=(1, 100), units='s')
RotationPhase._add_meta_data('duration_ref', val=1, units='s')
RotationPhase._add_meta_data('angle_lower', val=0.0, units='rad')  # rad
RotationPhase._add_meta_data('angle_upper', val=25 * np.pi / 180, units='rad')  # rad
RotationPhase._add_meta_data('angle_ref', val=1, units='rad')
RotationPhase._add_meta_data('angle_ref0', val=0, units='rad')
RotationPhase._add_meta_data('angle_defect_ref', val=0.01, units='rad')
RotationPhase._add_meta_data('velocity_lower', val=0, units='kn')
RotationPhase._add_meta_data('velocity_upper', val=1000, units='kn')
RotationPhase._add_meta_data('velocity_ref', val=100, units='kn')
RotationPhase._add_meta_data('velocity_ref0', val=0, units='kn')
RotationPhase._add_meta_data('velocity_defect_ref', val=None, units='kn')
RotationPhase._add_meta_data('mass_lower', val=0, units='lbm')
RotationPhase._add_meta_data('mass_upper', val=190_000, units='lbm')
RotationPhase._add_meta_data('mass_ref', val=100_000, units='lbm')
RotationPhase._add_meta_data('mass_ref0', val=0, units='lbm')
RotationPhase._add_meta_data('mass_defect_ref', val=None, units='lbm')
RotationPhase._add_meta_data('distance_lower', val=0, units='ft')
RotationPhase._add_meta_data('distance_upper', val=10.e3, units='ft')
RotationPhase._add_meta_data('distance_ref', val=3000, units='ft')
RotationPhase._add_meta_data('distance_ref0', val=0, units='ft')
RotationPhase._add_meta_data('distance_defect_ref', val=3000, units='ft')
RotationPhase._add_meta_data('normal_ref', val=1, units='lbf')
RotationPhase._add_meta_data('normal_ref0', val=0, units='lbf')
RotationPhase._add_meta_data('t_init_gear', val=100, units='s')
RotationPhase._add_meta_data('t_init_flaps', val=100, units='s')
RotationPhase._add_meta_data('num_segments', val=None, units='unitless')
RotationPhase._add_meta_data('order', val=None, units='unitless')

# Adding initial guess metadata
RotationPhase._add_initial_guess_meta_data(
    InitialGuessTime(),
    desc='initial guess for time options')
RotationPhase._add_initial_guess_meta_data(
    InitialGuessState('alpha'),
    desc='initial guess for angle of attack state')
RotationPhase._add_initial_guess_meta_data(
    InitialGuessState('velocity'),
    desc='initial guess for true airspeed state')
RotationPhase._add_initial_guess_meta_data(
    InitialGuessState('mass'),
    desc='initial guess for mass state')
RotationPhase._add_initial_guess_meta_data(
    InitialGuessState('distance'),
    desc='initial guess for distance state')
RotationPhase._add_initial_guess_meta_data(
    InitialGuessControl('throttle'),
    desc='initial guess for throttle')



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class CanardMass(om.ExplicitComponent):
    '''
    Calculates the mass of the canard. The methodology is based on the FLOPS weight
    equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Canard.AREA, val=0.0)
        add_aviary_input(self, Aircraft.Canard.TAPER_RATIO, val=00)
        add_aviary_input(self, Aircraft.Canard.MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.Canard.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        togw = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        area = inputs[Aircraft.Canard.AREA]
        taper_ratio = inputs[Aircraft.Canard.TAPER_RATIO]

        canard_weight = 0.53 * area * togw**0.2 * (taper_ratio+0.5)
        outputs[Aircraft.Canard.MASS] = canard_weight * \
            inputs[Aircraft.Canard.MASS_SCALER] / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        area = inputs[Aircraft.Canard.AREA]
        taper_ratio = inputs[Aircraft.Canard.TAPER_RATIO]
        scalar = inputs[Aircraft.Canard.MASS_SCALER]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM

        gross_weight_exp = gross_weight**0.2

        J[Aircraft.Canard.MASS, Aircraft.Canard.AREA] = 0.53 * \
            scalar * (taper_ratio+0.5) * gross_weight_exp / GRAV_ENGLISH_LBM

        J[Aircraft.Canard.MASS, Aircraft.Canard.TAPER_RATIO] = \
            0.53*area * scalar * gross_weight_exp / GRAV_ENGLISH_LBM

        J[Aircraft.Canard.MASS, Aircraft.Canard.MASS_SCALER] = \
            0.53*area * (taper_ratio+0.5) * gross_weight_exp / GRAV_ENGLISH_LBM

        J[Aircraft.Canard.MASS, Mission.Design.GROSS_MASS] = \
            (0.106*area * scalar * (taper_ratio+0.5)) / gross_weight**0.8



import os

import numpy as np
import openmdao.api as om

from aviary.variable_info.variable_meta_data import _MetaData

base_path = os.path.abspath(os.path.join(__file__, "..", "..", ".."))


def assert_no_duplicates(
    check_list,
):
    full_list = []
    for entry in check_list:
        if np.shape(entry) == ():
            full_list.append(entry)
        else:
            entry = set(entry)
            for name in entry:
                full_list.append(name)
    full_list = list(filter(filter_empty, full_list))

    sorted_list = sorted(full_list)
    set_list = set(full_list)
    duplicates = []
    if len(sorted_list) != len(set_list):
        for idx, check_name in enumerate(sorted_list[1:]):
            if check_name == sorted_list[idx]:
                duplicates.append(check_name)
        duplicates = list(set(duplicates))

        raise ValueError(
            f"The variables {duplicates} are duplicates in the provided list."
        )

    return duplicates


def assert_structure_alphabetization(variables_file_loc):
    #################################################################################################################
    # Extract Class and Variable Names from File #
    #################################################################################################################

    full_file_loc = os.path.abspath(os.path.join(base_path, variables_file_loc))
    with open(full_file_loc) as variable_file:
        # variable_file = open(full_file_loc, "r")
        lines = variable_file.readlines()
        variable_file.close()
        ct = 0
        superclass_names = []  # comprehensive list of the names of the superclasses
        subclass_names_tot = []  # comprehensive list of the names of the subclasses
        subclass_names = []  # list of names of subclasses in each individual superclass, intermediate variable that is overwritten for each superclass
        variable_names_tot = []  # comprehensive list of the names of the variables
        variable_names = []  # list of names of variable in each individual subclass, intermediate variable that is overwritten for each subclass
        # list of names of variables in each individual superclass, intermediate variable that is overwritten for each superclass
        variable_names_tot_per_super = []

        first_super = True
        first_sub = True

        for ct, line in enumerate(lines):
            line_ls = line.lstrip()
            if line_ls.startswith('#'):
                # Skip any comments.
                continue

            leading_spaces = len(line) - len(line_ls)

            # Get Class Names #
            is_class = line.find("class") != -1 and (line[line.find("class") + 5]) == " "

            if is_class:
                split_line = line.split("class ")
                class_name = split_line[1].split(":")
                class_name = class_name[0]  # name of class on this line
                if split_line[0] == "":  # this means it is a top level class
                    superclass_names.append(class_name)
                    if subclass_names:  # checks to see if any subclasses have been encountered yet
                        # appends the subclass names in the previous superclass to the total subclass list
                        subclass_names_tot.append(subclass_names)
                    subclass_names = []  # resets the list of subclasses in one individual superclass

                    if variable_names_tot_per_super:  # checks to see if any arrays of variables need to be added
                        if variable_names:  # checks to see if any lists of variable names need to be added
                            # appends the variable names in the previous subclass to the total variable names list per superclass
                            variable_names_tot_per_super.append(variable_names)
                        variable_names = []  # resets the list of variable names in one individual subclass
                        # appends the variable names of the entire previous superclass to the total variable names list
                        variable_names_tot.append(variable_names_tot_per_super)
                    variable_names_tot_per_super = []

                elif split_line[0] == "    ":  # this means it is a subclass
                    subclass_names.append(class_name)
                    if variable_names:  # checkts to see if any variables have been encountered yet in this superclass
                        # appends the variable names in the previous subclass to the total variable names list per superclass
                        variable_names_tot_per_super.append(variable_names)
                    variable_names = []  # resets the list of variable names in one individual subclass

            # Get Variable Names #
            if leading_spaces == 8:
                var_name = line.split()[0]
                # append variable to variable name list for variables in a single subclass
                variable_names.append(var_name)

        # Append final variable names #
        if subclass_names:
            subclass_names_tot.append(subclass_names)

        if variable_names:
            variable_names_tot_per_super.append(variable_names)

        if variable_names_tot_per_super:
            variable_names_tot.append(variable_names_tot_per_super)

        #################################################################################################################
        # Finish Extracting Class and Variable Names from File #
        #################################################################################################################

        # Check For Alphabetization in Superclasses #
        sorted_superclass = superclass_names.copy()
        sorted_superclass.sort(key=str.casefold)
        superclass_is_sorted = sorted_superclass == superclass_names
        if not superclass_is_sorted:
            raise ValueError(
                f'The list of classes in the "{variables_file_loc}" file is not alphabetical. The class order should be: {sorted_superclass} but instead it is: {superclass_names}.'
            )

        # Check for Alphabetization in Subclasses #
        for idx, family in enumerate(subclass_names_tot):
            sorted_family = family.copy()
            sorted_family.sort(key=str.casefold)
            family_is_sorted = sorted_family == family
            if not family_is_sorted:
                raise ValueError(
                    f'The list of classes in "{superclass_names[idx]}" is not alphabetical. The class order should be: {sorted_family} but instead it is: {family}.'
                )

        # Check for Alphabetization in Variables #
        for super_cnt, var_per_super in enumerate(variable_names_tot):
            for sub_cnt, var_per_sub in enumerate(var_per_super):
                sorted_var_list = var_per_sub.copy()
                sorted_var_list.sort(key=str.casefold)
                var_list_is_sorted = sorted_var_list == var_per_sub
                if not var_list_is_sorted:
                    raise ValueError(
                        f'The list of variables in class "{subclass_names_tot[super_cnt][sub_cnt]}" within class "{superclass_names[super_cnt]}" is not alphabetical. The variable order should be: {sorted_var_list} but instead it is: {var_per_sub}.'
                    )


def assert_metadata_alphabetization(metadata_variables_list):
    previous_var = metadata_variables_list[0].split(":")
    out_of_order_vars = []

    for var_name in metadata_variables_list[1:]:
        current_var = var_name.split(":")
        max_size = min(len(current_var), len(previous_var))
        same = [previous_var[ct] == current_var[ct] for ct in np.arange(max_size)]
        try:
            diff_idx = same.index(False)
        except ValueError:
            raise ValueError(
                "There are two variables in the metadata file that have the same string name."
            )
        old_to_new = [previous_var[diff_idx], current_var[diff_idx]]
        old_to_new_alphabetical = old_to_new.copy()
        old_to_new_alphabetical.sort(key=str.casefold)
        is_alphabetical = old_to_new == old_to_new_alphabetical
        previous_var = current_var

        if not is_alphabetical:
            out_of_order_vars.append(var_name)

    if out_of_order_vars:
        raise ValueError(
            f"The variable(s) {out_of_order_vars} in the metadata are out of alphabetical order with their previous value."
        )


def assert_match_varnames(system, MetaData=None):

    prob = om.Problem()
    prob.model = system
    prob.setup()
    prob.final_setup()
    sys_inputs = prob.model.list_inputs(out_stream=None, prom_name=True)
    sys_inputs = set([val[1]["prom_name"] for val in sys_inputs])
    sys_outputs = prob.model.list_outputs(out_stream=None, prom_name=True)
    sys_outputs = set([val[1]["prom_name"] for val in sys_outputs])

    proper_var_names = set(
        [key for key in (_MetaData if MetaData is None else MetaData)])

    input_overlap = sys_inputs.intersection(proper_var_names)
    output_overlap = sys_outputs.intersection(proper_var_names)

    if input_overlap != sys_inputs:
        diff = sys_inputs - input_overlap
        raise ValueError(
            f"The inputs {diff} in the provided subsystem are not found in the provided variable structure."
        )

    if output_overlap != sys_outputs:
        diff = sys_outputs - output_overlap
        raise ValueError(
            f"The outputs {diff} in the provided subsystem are not found in the provided variable structure."
        )


def get_names_from_hierarchy(hierarchy):

    names = []
    keys = vars(hierarchy).keys()  # get initial class keys
    # filter out keys that aren't for relevant variables
    keys = list(filter(filter_underscore, list(keys)))

    for key in keys:
        subclass_vars = vars(
            getattr(hierarchy, key)
        )  # grab dictionary of variables for the subclass
        # filter out keys that aren't for relevant variables
        subclass_keys = list(filter(filter_underscore, list(subclass_vars.keys())))

        for var_name in subclass_keys:
            names.append(
                subclass_vars[var_name]
            )  # add relevant variables to dictionary

    return names


class DuplicateHierarchy:

    class Design:
        CRUISE_MACH = 'mission:design:cruise_mach'
        RANGE = 'mission:design:range'

    class OperatingLimits:
        MAX_MACH = 'mission:design:cruise_mach'  # this is a duplicate


def filter_empty(entry):
    if (entry != '') and (entry != None):
        return True

    else:
        return False


def filter_underscore(entry):
    if '__' not in entry:
        return True

    else:
        return False



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.anti_icing import AntiIcingMass
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft


class AntiIcingMassTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            "anti_icing",
            AntiIcingMass(aviary_options=get_flops_inputs(case_name)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.AntiIcing.MASS_SCALER,
                        Aircraft.Fuselage.MAX_WIDTH,
                        Aircraft.Nacelle.AVG_DIAMETER,
                        Aircraft.Wing.SPAN,
                        Aircraft.Wing.SWEEP],
            output_keys=Aircraft.AntiIcing.MASS,
            tol=3.0e-3)

    def test_case_2(self):
        # test with more than four engines
        prob = self.prob

        aviary_options = get_flops_inputs('LargeSingleAisle1FLOPS')
        aviary_options.set_val(Aircraft.Engine.NUM_ENGINES, np.array([5]))
        aviary_options.set_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES, 5)

        prob.model.add_subsystem(
            "anti_icing",
            AntiIcingMass(aviary_options=aviary_options),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        prob.set_val(Aircraft.AntiIcing.MASS_SCALER, 1.0)
        prob.set_val(Aircraft.Fuselage.MAX_WIDTH, 12.33, 'ft')
        prob.set_val(Aircraft.Nacelle.AVG_DIAMETER, np.array([7.94]), 'ft')
        prob.set_val(Aircraft.Wing.SPAN, 117.83, 'ft')
        prob.set_val(Aircraft.Wing.SWEEP, 25.0, 'deg')

        prob.run_model()

        mass = prob.get_val(Aircraft.AntiIcing.MASS)
        expected_mass = 305.14673602

        assert_near_equal(mass, expected_mass, tolerance=1e-10)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-10)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.mass.gasp_based.fuel import (BodyTankCalculations,
                                                    FuelAndOEMOutputs,
                                                    FuelMass, FuelMassGroup,
                                                    FuelSysAndFullFuselageMass,
                                                    FuselageAndStructMass)
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft, Mission


# this is the large single aisle 1 V3 test case
class BodyCalculationTestCase1(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "wing_calcs", BodyTankCalculations(aviary_options=get_option_defaults(), ), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_DESIGN, val=857.480639944284, units="ft**3"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX, val=1114.006551379108, units="ft**3"
        )
        self.prob.model.set_input_defaults("fuel_mass_min", val=32853, units="lbm")
        self.prob.model.set_input_defaults(
            Mission.Design.FUEL_MASS_REQUIRED, val=42892.0, units="lbm")
        self.prob.model.set_input_defaults("max_wingfuel_mass", val=55725.1, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, val=1114.0, units="ft**3"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Mission.Design.FUEL_MASS, val=42893.1, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.OPERATING_MASS, val=96508, units="lbm"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 2e-4
        assert_near_equal(
            self.prob[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY], 0, tol
        )  # note: not in version 3 output, calulated by hand
        assert_near_equal(
            self.prob["extra_fuel_volume"], 0, tol
        )  # note: not in version 3 output, calulated by hand
        assert_near_equal(
            self.prob["max_extra_fuel_mass"], 0, tol
        )  # note: not in version 3 output, calulated by hand
        assert_near_equal(self.prob["wingfuel_mass_min"], 32853.0, tol)
        # note: Aircraft.Fuel.TOTAL_CAPACITY is calculated differently in V3, so it is not included here

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class BodyCalculationTestCase2(
    unittest.TestCase
):  # this is v 3.6 large single aisle 1 test case with wing loading of 150 psf and fuel margin of 10%
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "wing_calcs", BodyTankCalculations(aviary_options=get_option_defaults(), ), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_DESIGN, val=989.2, units="ft**3")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX, val=876.7, units="ft**3")
        self.prob.model.set_input_defaults("fuel_mass_min", val=34942.7, units="lbm")
        self.prob.model.set_input_defaults(
            Mission.Design.FUEL_MASS_REQUIRED, val=44982.7, units="lbm")
        self.prob.model.set_input_defaults("max_wingfuel_mass", val=43852.1, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, val=876.7, units="ft**3"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Mission.Design.FUEL_MASS, val=44973.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.OPERATING_MASS, val=94417, units="lbm"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 3e-4
        assert_near_equal(self.prob[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY], 1130.6, tol)
        assert_near_equal(self.prob["extra_fuel_volume"], 112.5, tol)
        assert_near_equal(self.prob["max_extra_fuel_mass"], 5628.9, tol)
        assert_near_equal(self.prob["wingfuel_mass_min"], 29313.9, tol)
        assert_near_equal(self.prob[Aircraft.Fuel.TOTAL_CAPACITY], 46093.7, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


# this is the large single aisle 1 V3 test case
class FuelAndOEMTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("wing_calcs", FuelAndOEMOutputs(
            aviary_options=get_option_defaults(), ), promotes=["*"])

        self.prob.model.set_input_defaults(
            Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Propulsion.MASS, val=16129, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.STRUCTURE_MASS, val=50461.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_EQUIPMENT_MASS, val=21089.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_USEFUL_LOAD, val=4932.0, units="lbm")
        self.prob.model.set_input_defaults(
            Mission.Design.FUEL_MASS_REQUIRED, val=42892.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, 1114, units="ft**3"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_MARGIN, val=0, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.TOTAL_CAPACITY, val=55725.1, units="lbm")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 2e-4
        assert_near_equal(self.prob["OEM_wingfuel_mass"], 78894, tol)
        assert_near_equal(
            self.prob[Aircraft.Fuel.WING_VOLUME_DESIGN], 857.480639944284, tol)
        assert_near_equal(self.prob["OEM_fuel_vol"], 1577.160566039489, tol)
        assert_near_equal(self.prob[Aircraft.Design.OPERATING_MASS], 96508.0, tol)

        assert_near_equal(self.prob["payload_mass_max_fuel"], 23166.9, tol)
        assert_near_equal(self.prob["volume_wingfuel_mass"], 55725.1, tol)
        assert_near_equal(self.prob["max_wingfuel_mass"], 55725.1, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-12, rtol=1e-12)


class FuelSysAndFullFusMassTestCase(
    unittest.TestCase
):  # this is the large single aisle 1 V3 test case
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "sys_and_fus", FuelSysAndFullFuselageMass(aviary_options=get_option_defaults(), ), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS, val=15830.0, units="lbm")
        self.prob.model.set_input_defaults(
            "wing_mounted_mass", val=24446.343040697346, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        self.prob.model.set_input_defaults(
            Mission.Design.FUEL_MASS, val=42893, units="lbm")
        self.prob.model.set_input_defaults("wingfuel_mass_min", val=32853, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_MARGIN, val=0, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(self.prob["fus_mass_full"], 102270, tol)
        assert_near_equal(self.prob[Aircraft.Fuel.FUEL_SYSTEM_MASS], 1759, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


# this is the large single aisle 1 V3 test case
class FusAndStructMassTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "fus_and_struct", FuselageAndStructMass(aviary_options=get_option_defaults(), ), promotes=["*"]
        )

        self.prob.model.set_input_defaults("fus_mass_full", val=102270, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.MASS_COEFFICIENT, val=128, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.WETTED_AREA, val=4000, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.TailBoom.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults("pylon_len", val=0, units="ft")
        self.prob.model.set_input_defaults("min_dive_vel", val=420, units="kn")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.893, units="unitless")
        self.prob.model.set_input_defaults("MAT", val=0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.MASS, val=15830, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS, val=2275, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS, val=2297, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.POD_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS, val=3785, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.STRUCTURAL_MASS_INCREMENT, val=0, units="lbm")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(self.prob[Aircraft.Design.STRUCTURE_MASS], 50461.0, tol)
        assert_near_equal(self.prob[Aircraft.Fuselage.MASS], 18763, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=4e-12, rtol=1e-12)


class FuelMassTestCase(unittest.TestCase):  # this is the large single aisle 1 V3 test case
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("fuel", FuelMass(
            aviary_options=get_option_defaults(), ), promotes=["*"])

        self.prob.model.set_input_defaults(
            Aircraft.Design.STRUCTURE_MASS, val=50461.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS, val=1759, units="lbm")
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults("eng_comb_mass", val=14370.8, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_EQUIPMENT_MASS, val=21089.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_USEFUL_LOAD, val=4932.0, units="lbm")
        self.prob.model.set_input_defaults("payload_mass_des", val=36000, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        self.prob.model.set_input_defaults(
            Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS, val=36000, units="lbm"
        )
        self.prob.model.set_input_defaults("payload_mass_max", val=46040, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_MARGIN, val=0, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(self.prob[Mission.Design.FUEL_MASS], 42893, tol)
        assert_near_equal(self.prob[Aircraft.Propulsion.MASS], 16129, tol)
        assert_near_equal(self.prob[Mission.Design.FUEL_MASS_REQUIRED], 42892.0, tol)
        assert_near_equal(self.prob["fuel_mass_min"], 32853, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


# this is the large single aisle 1 V3 test case
class FuelMassGroupTestCase1(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("group", FuelMassGroup(
            aviary_options=get_option_defaults(), ), promotes=["*"])

        # top level
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.MASS, val=15830, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_EQUIPMENT_MASS, val=21089.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_USEFUL_LOAD, val=4932, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, 1114.0, units="ft**3"
        )

        # sys and fus
        self.prob.model.set_input_defaults(
            "wing_mounted_mass", val=24446.343040697346, units="lbm"
        )

        # fus and struct
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.MASS_COEFFICIENT, val=128, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.WETTED_AREA, val=4000, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.TailBoom.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults(
            "fus_and_struct.pylon_len", val=0, units="ft"
        )
        self.prob.model.set_input_defaults("min_dive_vel", val=420, units="kn")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.893, units="unitless")
        self.prob.model.set_input_defaults("fus_and_struct.MAT", val=0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS, val=2275, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS, val=2297, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.POD_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS, val=3785, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.STRUCTURAL_MASS_INCREMENT, val=0, units="lbm"
        )

        # fuel
        self.prob.model.set_input_defaults("eng_comb_mass", val=14370.8, units="lbm")
        self.prob.model.set_input_defaults("payload_mass_des", val=36000, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS, val=36000, units="lbm"
        )
        self.prob.model.set_input_defaults("payload_mass_max", val=46040, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_MARGIN, val=0, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4

        # wingfuel
        assert_near_equal(self.prob["fuel_and_oem.OEM_wingfuel_mass"], 78894, tol)
        assert_near_equal(
            self.prob["fuel_and_oem.OEM_fuel_vol"], 1577.160566039489, tol
        )
        assert_near_equal(self.prob[Aircraft.Design.OPERATING_MASS], 96508.0, tol)
        assert_near_equal(
            self.prob["fuel_and_oem.payload_mass_max_fuel"], 36000, tol
        )  # note: this is calculated differently in V3, so this is the V3.6 value
        assert_near_equal(self.prob["fuel_and_oem.volume_wingfuel_mass"], 55725.1, tol)
        assert_near_equal(self.prob["max_wingfuel_mass"], 55725.1, tol)

        # sys and fus
        assert_near_equal(self.prob["fus_mass_full"], 102270, tol)
        assert_near_equal(self.prob[Aircraft.Fuel.FUEL_SYSTEM_MASS], 1759, tol)

        # fus and struct
        assert_near_equal(self.prob[Aircraft.Design.STRUCTURE_MASS], 50461.0, tol)
        assert_near_equal(self.prob[Aircraft.Fuselage.MASS], 18763, tol)

        # fuel
        assert_near_equal(self.prob[Mission.Design.FUEL_MASS], 42893, tol)
        assert_near_equal(self.prob[Aircraft.Propulsion.MASS], 16129, tol)
        assert_near_equal(self.prob[Mission.Design.FUEL_MASS_REQUIRED], 42892.0, tol)
        assert_near_equal(self.prob["fuel_mass_min"], 32853, tol)

        # body tank
        assert_near_equal(
            self.prob[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY], 0, tol
        )  # note: not in version 3 output, calulated by hand
        assert_near_equal(
            self.prob["body_tank.extra_fuel_volume"], 0, tol
        )  # note: not in version 3 output, calulated by hand
        assert_near_equal(
            self.prob["body_tank.max_extra_fuel_mass"], 0, tol
        )  # note: not in version 3 output, calulated by hand
        # note: Aircraft.Fuel.TOTAL_CAPACITY is calculated differently in V3, so it is not included here

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-11, rtol=1e-12)


class FuelMassGroupTestCase2(
    unittest.TestCase
):  # this is v 3.6 large single aisle 1 test case with wing loading of 150 psf and fuel margin of 10%
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("group", FuelMassGroup(
            aviary_options=get_option_defaults(), ), promotes=["*"])

        # top level
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.MASS, val=13833, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3632, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_EQUIPMENT_MASS, val=21031.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_USEFUL_LOAD, val=4885.18, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, 876.7, units="ft**3"
        )

        # sys and fus
        self.prob.model.set_input_defaults(
            "wing_mounted_mass", val=24446.343040697346, units="lbm"
        )  # note: calculated by hand

        # fus and struct
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.MASS_COEFFICIENT, val=128, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.WETTED_AREA, val=4000, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.TailBoom.LENGTH, val=129.5, units="ft"
        )  # note: calculated by hand
        self.prob.model.set_input_defaults(
            "fus_and_struct.pylon_len", val=0, units="ft"
        )  # note: calculated by hand
        self.prob.model.set_input_defaults("min_dive_vel", val=420, units="kn")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.75, units="unitless")
        self.prob.model.set_input_defaults("fus_and_struct.MAT", val=0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS, val=2181, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS, val=2158, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.POD_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS, val=3785, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.STRUCTURAL_MASS_INCREMENT, val=0, units="lbm"
        )

        # fuel
        self.prob.model.set_input_defaults("eng_comb_mass", val=14370.8, units="lbm")
        self.prob.model.set_input_defaults("payload_mass_des", val=36000, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS, val=36000, units="lbm"
        )
        self.prob.model.set_input_defaults("payload_mass_max", val=46040, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_MARGIN, val=10, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4

        # wingfuel
        assert_near_equal(self.prob["fuel_and_oem.OEM_wingfuel_mass"], 80982.7, tol)
        assert_near_equal(self.prob["fuel_and_oem.OEM_fuel_vol"], 1618.9, tol)
        assert_near_equal(self.prob[Aircraft.Design.OPERATING_MASS], 94417.0, tol)
        assert_near_equal(self.prob["fuel_and_oem.payload_mass_max_fuel"], 34879.2, tol)
        assert_near_equal(self.prob["fuel_and_oem.volume_wingfuel_mass"], 43852.1, tol)
        assert_near_equal(self.prob["max_wingfuel_mass"], 43852.1, tol)

        # sys and fus
        assert_near_equal(
            self.prob["fus_mass_full"], 107806.75695930266, tol
        )  # note: calculated by hand
        assert_near_equal(self.prob[Aircraft.Fuel.FUEL_SYSTEM_MASS], 2029, tol)

        # fus and struct
        assert_near_equal(self.prob[Aircraft.Design.STRUCTURE_MASS], 48470.0, tol)
        assert_near_equal(
            self.prob[Aircraft.Fuselage.MASS], 19002.0, 0.00054
        )  # tol is slightly higher because GASP iteration is less rigorous.

        # fuel
        assert_near_equal(self.prob[Mission.Design.FUEL_MASS], 44982.7, tol)
        assert_near_equal(self.prob[Aircraft.Propulsion.MASS], 16399.0, tol)
        assert_near_equal(self.prob[Mission.Design.FUEL_MASS_REQUIRED], 44982.7, tol)
        assert_near_equal(self.prob["fuel_mass_min"], 34942.7, tol)

        # body tank
        assert_near_equal(
            self.prob[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY], 1130.6, 0.0112
        )  # tol is slightly higher because GASP iteration is less rigorous, and also because of numerical issues with inputs.
        assert_near_equal(
            self.prob["body_tank.extra_fuel_volume"], 112.5, 0.0022
        )  # tol is slightly higher because GASP iteration is less rigorous.
        assert_near_equal(
            self.prob["body_tank.max_extra_fuel_mass"], 5628.9, 0.0025
        )  # tol is slightly higher because GASP iteration is less rigorous.

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-11, rtol=1e-12)


class FuelMassGroupTestCase3(
    unittest.TestCase
):  # this is v 3.6 advanced tube and wing case
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("group", FuelMassGroup(
            aviary_options=get_option_defaults(), ), promotes=["*"])

        # top level
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=145388.0, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS, val=15098.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3765, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_EQUIPMENT_MASS, val=17201.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.FIXED_USEFUL_LOAD, val=4701, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, 677.6, units="ft**3"
        )

        # sys and fus
        self.prob.model.set_input_defaults(
            "wing_mounted_mass", val=17272.898624554255, units="lbm"
        )  # note: calculated by hand

        # fus and struct
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.MASS_COEFFICIENT, val=128, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.WETTED_AREA, val=4209, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.TailBoom.LENGTH, val=119.03, units="ft"
        )
        self.prob.model.set_input_defaults(
            "fus_and_struct.pylon_len", val=0, units="ft"
        )
        self.prob.model.set_input_defaults("min_dive_vel", val=420, units="kn")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=4.484, units="unitless")
        self.prob.model.set_input_defaults("fus_and_struct.MAT", val=0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS, val=2018, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS, val=1500, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=6140, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.POD_MASS_SCALER, val=1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS, val=2795, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Design.STRUCTURAL_MASS_INCREMENT, val=0, units="lbm"
        )

        # fuel
        self.prob.model.set_input_defaults("eng_comb_mass", val=9328.2, units="lbm")
        self.prob.model.set_input_defaults("payload_mass_des", val=30800, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS, val=30800, units="lbm"
        )
        self.prob.model.set_input_defaults("payload_mass_max", val=46770.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.FUEL_MARGIN, val=10, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4

        # wingfuel
        assert_near_equal(self.prob["fuel_and_oem.OEM_wingfuel_mass"], 62427.2, tol)
        assert_near_equal(self.prob["fuel_and_oem.OEM_fuel_vol"], 1248.0, tol)
        assert_near_equal(self.prob[Aircraft.Design.OPERATING_MASS], 82961.0, tol)
        assert_near_equal(self.prob["fuel_and_oem.volume_wingfuel_mass"], 33892.8, tol)
        assert_near_equal(self.prob["max_wingfuel_mass"], 33892.8, tol)

        # sys and fus
        assert_near_equal(self.prob[Aircraft.Fuel.FUEL_SYSTEM_MASS], 1426, tol)

        # fus and struct
        assert_near_equal(self.prob[Aircraft.Design.STRUCTURE_MASS], 46539.0, tol)
        assert_near_equal(self.prob[Aircraft.Fuselage.MASS], 18988, tol)

        # fuel
        assert_near_equal(self.prob[Mission.Design.FUEL_MASS], 31627.2, tol)
        assert_near_equal(self.prob[Aircraft.Propulsion.MASS], 10755.0, tol)
        assert_near_equal(self.prob[Mission.Design.FUEL_MASS_REQUIRED], 31627, tol)
        assert_near_equal(self.prob["fuel_mass_min"], 15657.2, tol)

        # body tank
        assert_near_equal(self.prob[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY], 0, tol)
        assert_near_equal(self.prob["body_tank.extra_fuel_volume"], 17.9, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=3e-9, rtol=6e-11)


if __name__ == "__main__":
    unittest.main()



from packaging import version
import warnings

import openmdao
import openmdao.api as om

from aviary.variable_info.variables import Aircraft, Mission
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables_in import VariablesIn
from aviary.variable_info.variable_meta_data import _MetaData
from aviary.variable_info.functions import override_aviary_vars

use_new_openmdao_syntax = version.parse(openmdao.__version__) >= version.parse("3.28")


class CorePreMission(om.Group):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )
        self.options.declare(
            'subsystems', desc='list of core subsystem builders'
        )
        self.options.declare(
            'meta_data', desc='problem metadata', default=_MetaData
        )
        # NOTE this flag is only needed for tests - in AviaryProblem it should always be False
        self.options.declare('process_overrides', types=bool, default=True,
                             desc='When True, overrides are handled here, otherwise, '
                                  'they are handled in the parent system.')

    def setup(self, **kwargs):
        if use_new_openmdao_syntax:
            # rely on openMDAO's auto-ordering for this group
            self.options['auto_order'] = True

        aviary_options = self.options['aviary_options']
        core_subsystems = self.options['subsystems']

        for subsystem in core_subsystems:
            self.add_subsystem(
                subsystem.name,
                subsystem.build_pre_mission(aviary_options),
                promotes_inputs=['*'],
                promotes_outputs=['*']
            )

    def configure(self):
        self.manual_overrides = []
        for subsystem in self.options['subsystems']:
            try:
                self.manual_overrides.extend(getattr(getattr(self, subsystem.name),
                                                     'manual_overrides'))
            except:
                continue

        if self.options['process_overrides']:
            override_aviary_vars(self, self.options["aviary_options"],
                                 manual_overrides=self.manual_overrides)



def add_meta_data(
        key,
        meta_data,
        units='unitless',
        desc=None,
        default_value=0.0,
        option=False,
        types=None,
        historical_name=None,
        _check_unique=True):
    '''
    Add new meta data associated with variables in the Aviary data hierarchy.

    Parameters
    ----------
    key : str
        Aviary variable name

    meta_data : dict
        dictionary of meta data to add the variable to

    units : str or None
        units of measure

    desc : str
        brief description of the variable

    default_value : any
        in context, the Aviary value assumed if the variable is missing from
        options and/or inputs

        Note, a default value of `None` indicates that the variable is
        optional, but that there is no default.

    option : bool
        indicates that this variable is an option, rather than a normal input

    types : type
        gives the allowable type(s) of the variable

    historical_name : dict or None
        dictionary of names that the variable held in prior codes

        Example: {"FLOPS":"WTIN.WNGWT", "LEAPS1": "aircraft.inputs.wing_weight", "GASP": "INGASP.WWGHT"}

        NAMELIST nameing convention
            &<function_name>.<namelist_name>.<var_name>

            Example: &DEFINE.CONFIN.GW
                represents the GW variable of the CONFIN namelist as defined in
                the DEFINE subroutine

        COMMON block naming convention, including aliases:
            <block_name>.<var_name>

            Example: CONFIG.GW
                represents the GW variable of the CONFIG common block

        Local variable naming convention, including equivalence statements, parameters, and other local declarations:
            ~<function_name>.<var_name>

            Example: ~ANALYS.GWTOL
                represents the GWTOL variable of the ANALYS subroutine

    _check_unique : bool
        private use only flag that tells whether to check the meta_data for the pre-existing presence
        of the provided key. This should only be set to false when update_meta_data is the calling function.

    Returns
    ------- 
    None
        No variables returned by this method.

    Raises
    ----------
    None
        No exceptions raised by this method, although other methods called within may raise exceptions.
    '''

    if key in meta_data and _check_unique:
        raise ValueError(
            f'You added the variable {key} to a variable metadata dictionary via the add_meta_data function, but {key} already was present in the dictionary. If you are sure you want to overwrite this variable, call the update_meta_data function instead.')

    if units is None:
        units = 'unitless'

    meta_data[key] = {
        'historical_name': historical_name,
        'units': units,
        'desc': desc,
        'option': option,
        'default_value': default_value,
        'types': types
    }


def update_meta_data(
        key,
        meta_data,
        units='unitless',
        desc=None,
        default_value=0.0,
        option=False,
        types=None,
        historical_name=None):
    '''
    Update existing meta data associated with variables in the Aviary data hierarchy.

    Parameters
    ----------
    key : str
        Aviary variable name

    meta_data : dict
        dictionary of meta data to add the variable to

    units : str or None
        units of measure

    desc : str
        brief description of the variable

    default_value : Any
        in context, the Aviary value assumed if the variable is missing from
        options and/or inputs

        Note, a default value of `None` indicates that the variable is
        optional, but that there is no default.

    option : bool
        indicates that this variable is an option, rather than a normal input

    types : type
        gives the allowable type(s) of the variable

    historical_name : dict or None
        dictionary of names that the variable held in prior codes

        Example: {"FLOPS":"WTIN.WNGWT", "LEAPS1": "aircraft.inputs.wing_weight", "GASP": "INGASP.WWGHT"}

        NAMELIST nameing convention
            &<function_name>.<namelist_name>.<var_name>

            Example: &DEFINE.CONFIN.GW
                represents the GW variable of the CONFIN namelist as defined in
                the DEFINE subroutine

        COMMON block naming convention, including aliases:
            <block_name>.<var_name>

            Example: CONFIG.GW
                represents the GW variable of the CONFIG common block

        Local variable naming convention, including equivalence statements, parameters, and other local declarations:
            ~<function_name>.<var_name>

            Example: ~ANALYS.GWTOL
                represents the GWTOL variable of the ANALYS subroutine

    Returns
    ------- 
    None
        No variables returned by this method.

    Raises
    ----------
    None
        No exceptions raised by this method, although other methods called within may raise exceptions.
    '''

    if key not in meta_data:
        raise ValueError(
            f'You provided the variable {key} to a variable metadata dictionary via the update_meta_data function, but {key} does not exist in the dictionary. If you are sure you want to add this variable to the dictionary, call the add_meta_data function instead.')

    add_meta_data(key=key, meta_data=meta_data, units=units, desc=desc,
                  default_value=default_value, option=option, types=types, historical_name=historical_name, _check_unique=False)



import getpass
import numpy as np
import re
import warnings

from datetime import datetime
from pathlib import Path

from openmdao.utils.units import valid_units, is_compatible

from aviary.utils.named_values import get_items, get_keys
from aviary.utils.functions import get_path
from aviary.utils.named_values import NamedValues


def read_data_file(filename: (str, Path), metadata=None, aliases=None,
                   save_comments=False):
    """
    Read data file in Aviary format, which is data delimited by commas with any amount of
    whitespace allowed between data entries. Spaces are not allowed in openMDAO
    variables, so any spaces in header entries are replaced with underscores.

    Parameters
    ----------
    filename : (str, Path)
        filename or filepath of data file to be read
    metadata : dict, optional
        metadata to check validity of variable names provided in data file. Columns with
        variable names that can't be found in metadata will be skipped. If not provided,
        all validly formatted columns are always read.
    aliases : dict, optional
        optional dictionary to define a mapping of variables to allowable aliases in the
        data file header. Keys are variable names, to be used in openMDAO, values
        are a list of headers that correspond to that variable. Alias matching is not
        case-sensitive, and underscores and spaces are treated as equivalent.
    save_comments : bool, optional
        flag if comments in data file should be returned along with data. Defaults to 
        False.

    Returns
    -------
    data : NamedValues
        data read from file in NamedValues format, including variable name, units, and 
        values (stored in a numpy array)
    comments : list of str
        any comments from file, with comment characters ('#') stripped out (only if 
        save_comments=True)
    """
    filepath = get_path(filename)

    data = NamedValues()
    comments = []

    # prep aliases for case-insensitive matching, with spaces == underscores
    if aliases:
        for key in aliases:
            if isinstance(aliases[key], str):
                aliases[key] = [aliases[key]]
            aliases[key] = [re.sub('\s', '_', item).lower() for item in aliases[key]]

    with open(filepath, newline=None, encoding='utf-8-sig') as file:
        # csv.reader() and other avaliable packages that can read csv files are not used
        # Manual control of file reading ensures that comments are kept intact and other
        # checks can be performed
        check_for_header = True
        for line_count, line_data in enumerate(file):
            # if comments are present in line, strip them out
            if '#' in line_data:
                index = line_data.index('#')
                comments.append(line_data[index+1:].strip())
                line_data = line_data[:index]

            # split by delimiters, remove whitespace and newline characters
            line_data = re.split(r'[;,]\s*', line_data.strip())

            # ignore empty lines
            if not line_data or line_data == ['']:
                continue

            # try to convert line_data to float, skip any blank strings
            try:
                line_data = [float(var) for var in line_data if var != '']
            # data contains things other than floats
            except (ValueError):
                # skip checking for header data if not required
                if check_for_header:
                    # dictionary of header name: units
                    header = {}
                    # list of which column goes with each valid header entry
                    valid_indices = []
                    for index in range(len(line_data)):
                        item = re.split('[(]', line_data[index])
                        item = [item[i].strip(') ') for i in range(len(item))]
                        # openMDAO vars can't have spaces, convert to underscores
                        name = re.sub('\s', '_', item[0])
                        if aliases:
                            # "reverse" lookup name in alias dict
                            for key in aliases:
                                if name.lower() in aliases[key]:
                                    name = key
                                    break
                        # 'default' default_units
                        default_units = 'unitless'
                        # if metadata is provided, ensure variable exists and update
                        # default_units
                        if metadata is not None:
                            if name not in metadata.keys():
                                warnings.warn(f'Header <{name}> was not recognized, and '
                                              'will be skipped'
                                              )
                                continue
                            else:
                                default_units = metadata[name]['units']

                        # if units are provided, check that they are valid
                        if len(item) > 1:
                            units = item[-1]
                            if valid_units(item[1]):
                                # check that units are compatible with expected units
                                if metadata is not None:
                                    if not is_compatible(units, default_units):
                                        # Raising error here, as trying to use default
                                        # units could mean accidental conversion which
                                        # would significantly impact analysis
                                        raise ValueError(f'Provided units of <{units}> '
                                                         f'for column <{name}>, which '
                                                         'are not compatible with default '
                                                         f'units of {default_units}')
                            else:
                                # Units were not recognized. Raise error
                                raise ValueError(f'Invalid units <{units}> provided for '
                                                 f'column <{name}> while reading '
                                                 f'<{filepath}>.')
                        else:
                            if metadata is not None and default_units != 'unitless':
                                # units were not provided, but variable should have them
                                # assume default units for that variable
                                warning = f'Units were not provided for column <{name}> '\
                                          f'while reading <{filepath}>. Using default '\
                                          f'units of {default_units}.'
                                warnings.warn(warning)
                            units = default_units

                        header[name] = units
                        valid_indices.append(index)

                    if len(header) > 0:
                        check_for_header = False
                        raw_data = {key: [] for key in header.keys()}
                        continue

                # only raise error if not checking for header, or invalid header found
                raise ValueError(
                    f'Non-numerical value found in data file <{filepath}> on line '
                    f'{str(line_count)}')

            # This point is reached when the first valid numerical entry in data file
            # is found. Stop looking for header data from now on
            check_for_header = False

            # pull out data for each valid header, ignore other columns
            for idx, variable in enumerate(header.keys()):
                # valid_indices matches dictionary order, pull data from correct column
                raw_data[variable].append(line_data[valid_indices[idx]])

    # store data in NamedValues object
    for variable in header.keys():
        data.set_val(variable, val=np.array(raw_data[variable]), units=header[variable])

    if save_comments:
        return data, comments
    else:
        return data


def write_data_file(filename: (str, Path) = None, data: NamedValues = None,
                    comments: (str, list) = [], include_timestamp: bool = False):
    """
    Write data to a comma-separated values (csv) format file using the Aviary data table
    format.

    Parameters
    ----------
    filename : (str, Path)
        filename or filepath for data file to be written
    data : NamedValues
        NamedValues object containing data that will be written to file, which includes
        variable name, units, and values
    comments : (str, list of str), optional
        optional comments that will be included in the top of the output file, before
        data begins
    include_timestamp : bool, optional
        optional flag to set if timestamp and user should be include in file comments
    """
    if isinstance(filename, str):
        filepath = Path(filename)
    else:
        filepath = filename

    if data is None:
        raise UserWarning(f'No data provided to write to {filepath.name}')

    if type(comments) is str:
        comments = [comments]

    # strip '#' from comments - np.savetxt() will automatically add them
    for idx, line in enumerate(comments):
        if len(line) > 0:
            if line[0] != '#':
                comments[idx] = '# ' + line.strip()

    # if there are comments, add some spacing afterwards - otherwise it should be empty
    if comments:
        comments.append('\n')

    if include_timestamp:
        timestamp = datetime.now().strftime('%m/%d/%y at %H:%M')
        try:
            user = ' by ' + getpass.getuser()
        except Exception:
            user = ''
        stamp = [f'# created {timestamp}{user}\n']
        comments = stamp + comments

    # assemble separate variable name and units information into single list for header
    header = []
    data_dict = {}
    for var, val_and_units in get_items(data):
        units = val_and_units[1]
        formatted_units = ''
        # only explicitly include units if there are any
        if units is not None and units != 'unitless':
            formatted_units = ' (' + units + ')'
        header.append(var + formatted_units)
        data_dict[var] = np.array([str(i) for i in data.get_val(var, units)])

    # set column widths, for more human-readable format
    col_format = []
    for i, key in enumerate(get_keys(data)):
        header_len = len(header[i])
        data_len = len(max(data_dict[key], key=len))
        # min column width is 10 - spaced out columns are visually easier to follow
        # don't pad first column
        if i > 0:
            min_width = 10
        else:
            min_width = 0
        col_len = max(header_len, data_len, min_width)

        # if headers are smaller than column, pad with leading whitespace
        if header_len < col_len:
            header[i] = ' '*(col_len-header_len) + header[i]

        # special string to define column formatting with specific width
        format = f'%{col_len}s'
        # don't include commas for last column
        if i < len(header)-1:
            format = format + ', '
        col_format.append(format)

    # convert engine_data from dict to array so it can be written using savetxt
    formatted_data = np.array([data_dict[key] for key in data_dict]).transpose()

    # write to output file w/ header and comments
    np.savetxt(filepath, formatted_data,
               fmt=''.join(col_format),
               delimiter=',',
               header=', '.join(header),
               comments='\n'.join(comments))
